""" Quantum Inspire SDK

This file contains code modified from https://github.com/ProjectQ-Framework/ProjectQ in the QIBackend class.
The ProjectQ code is under the Apache License 2.0.


Copyright 2018 QuTech Delft

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import random
from collections import defaultdict
from functools import reduce
from typing import List, Dict, Iterator, Union, Optional, Any, Set

from projectq.cengines import BasicEngine
from projectq.meta import LogicalQubitIDTag, get_control_count
from projectq.ops import (NOT, Allocate, Barrier, Deallocate, FlushGate, H,
                          Measure, Ph, Rx, Ry, Rz, S, Sdag, Swap, T, Tdag, X,
                          Y, Z, Command)
from projectq.types import Qubit
from quantuminspire.exceptions import AuthenticationError
from quantuminspire.api import QuantumInspireAPI
from quantuminspire.exceptions import ProjectQBackendError


class QIBackend(BasicEngine):  # type: ignore
    """ Backend for Quantum Inspire

    """

    def __init__(self, num_runs: int = 1024, verbose: int = 0, quantum_inspire_api: Optional[QuantumInspireAPI] = None,
                 backend_type: Optional[Union[Dict[str, Any], int, str]] = None) -> None:
        """
        Initialize the Backend object.

        Args:
            num_runs: Number of runs to collect statistics (default is 1024).
            verbose: Verbosity level, defaults to 0, which produces no extra output.
            quantum_inspire_api: Connection to QI platform, optional parameter.
            backend_type: Backend to use for execution. When no backend_type is provided, the default backend will be
                          used.
        """
        BasicEngine.__init__(self)
        self._flushed: bool = False
        """ Because engines are meant to be 'single use' by the way ProjectQ is designed,
        any additional gates received after a FlushGate triggers an exception. """
        self._reset()
        self._num_runs: int = num_runs
        self._verbose: int = verbose
        self._cqasm: str = str()
        self._measured_states: Dict[int, float] = {}
        self._measured_ids: List[int] = []
        self._allocated_qubits: Set[int] = set()
        self._max_qubit_id: int = -1
        if quantum_inspire_api is None:
            try:
                quantum_inspire_api = QuantumInspireAPI()
            except AuthenticationError as ex:
                raise AuthenticationError('Make sure you have saved your token credentials on disk or '
                                          'provide a QuantumInspireAPI instance as parameter to QIBackend') from ex
        self.quantum_inspire_api: QuantumInspireAPI = quantum_inspire_api
        self.backend_type: Optional[Union[Dict[str, Any], int, str]] = backend_type

    def cqasm(self) -> str:
        """ Return cqasm code that is generated last. """
        return self._cqasm

    def is_available(self, cmd: Command) -> bool:
        """
        Return true if the command can be executed.

        Args:
            cmd: Command for which to check availability.
        """
        count = get_control_count(cmd)
        g = cmd.gate
        if self._verbose:
            print('call to is_available with cmd %s (gate %s)' % (cmd, g))
        if g == NOT and count <= 2:
            return True
        if g == Z and count <= 1:
            return True
        if g in (Measure, Allocate, Deallocate, Barrier):
            return True
        if count != 0:
            return False
        if g in (T, Tdag, S, Sdag, Swap, H, X, Y, Z):
            return True
        elif isinstance(g, (Rx, Ry, Rz)):
            return True
        elif isinstance(g, Ph):
            return False
        else:
            return False

    def _reset(self) -> None:
        """ Reset temporary variable qasm to an initial value and set a flag to clear variables in _store
            when _store is called. """
        self._clear = True
        self.qasm = ""

    def _store(self, cmd: Command) -> None:
        """
        Temporarily store the command cmd.

        Translates the command and stores the results in local variables.

        Args:
            cmd: Command to store.
        """
        if self._verbose >= 2:
            print('_store {0}: cmd {1}'.format(id(self), cmd))
            print('   _allocated_qubits {0}'.format(self._allocated_qubits))

        if self._clear:
            self._clear = False
            self.qasm = ""
            self._measured_states = {}
            self._measured_ids = []

        gate = cmd.gate

        if gate == Deallocate:
            index_to_remove = cmd.qubits[0][0].id
            self._allocated_qubits.discard(index_to_remove)
            if self._verbose >= 2:
                print('_store: Deallocate gate {0}'.format((index_to_remove,)))
                print('   _allocated_qubits {0}'.format(self._allocated_qubits))
            return

        if self._flushed:
            raise RuntimeError("Same instance of QIBackend used for circuit after Flush.")

        if gate == Allocate:
            index_to_add = cmd.qubits[0][0].id
            self._allocated_qubits.add(index_to_add)
            self._max_qubit_id = max(self._max_qubit_id, index_to_add)
            if self._verbose >= 2:
                print('_store: Allocate gate {0}'.format((index_to_add,)))
                print('   _allocated_qubits {0}'.format(self._allocated_qubits))
            return

        if gate == Measure:
            assert len(cmd.qubits) == 1 and len(cmd.qubits[0]) == 1
            logical_id = None
            for t in cmd.tags:
                if isinstance(t, LogicalQubitIDTag):
                    logical_id = t.logical_qubit_id
                    break
            if self.main_engine.mapper is None:
                logical_id = cmd.qubits[0][0].id  # no mapping
            assert logical_id is not None
            self._measured_ids += [logical_id]
        elif gate == NOT and get_control_count(cmd) == 1:
            # this case also covers the CX controlled gate
            ctrl_pos = cmd.control_qubits[0].id
            qb_pos = cmd.qubits[0][0].id
            self.qasm += "\ncnot q[{}], q[{}]".format(ctrl_pos, qb_pos)
        elif gate == Swap:
            q0 = cmd.qubits[0][0].id
            q1 = cmd.qubits[1][0].id
            self.qasm += "\nswap q[{}], q[{}]".format(q0, q1)
        elif gate == X and get_control_count(cmd) == 2:
            ctrl_pos1 = cmd.control_qubits[0].id
            ctrl_pos2 = cmd.control_qubits[1].id
            qb_pos = cmd.qubits[0][0].id
            self.qasm += "\ntoffoli q[{}], q[{}], q[{}]".format(ctrl_pos1, ctrl_pos2, qb_pos)
        elif gate == Z and get_control_count(cmd) == 1:
            ctrl_pos = cmd.control_qubits[0].id
            qb_pos = cmd.qubits[0][0].id
            self.qasm += "\ncz q[{}], q[{}]".format(ctrl_pos, qb_pos)
        elif gate == Barrier:
            qb_pos = [qb.id for qr in cmd.qubits for qb in qr]
            self.qasm += "\n# barrier gate "
            qb_str = ""
            for pos in qb_pos:
                qb_str += "q[{}], ".format(pos)
            self.qasm += qb_str[:-2] + ";"
        elif isinstance(gate, Rz) and get_control_count(cmd) == 1:
            ctrl_pos = cmd.control_qubits[0].id
            qb_pos = cmd.qubits[0][0].id
            gate_name = 'cr'
            self.qasm += "\n{} q[{}],q[{}],{:.12f}".format(gate_name, ctrl_pos, qb_pos, gate.angle)
        elif isinstance(gate, (Rx, Ry)) and get_control_count(cmd) == 1:
            raise NotImplementedError('controlled Rx or Ry gate not implemented')
        elif isinstance(gate, (Rx, Ry, Rz)):
            assert get_control_count(cmd) == 0
            qb_pos = cmd.qubits[0][0].id
            gate_name = str(gate)[0:2].lower()
            self.qasm += "\n{} q[{}],{:.12g}".format(gate_name, qb_pos, gate.angle)
        elif gate == Tdag and get_control_count(cmd) == 0:
            qb_pos = cmd.qubits[0][0].id
            self.qasm += "\ntdag q[{}]".format(qb_pos)
        elif gate == Sdag and get_control_count(cmd) == 0:
            qb_pos = cmd.qubits[0][0].id
            self.qasm += "\nsdag q[{}]".format(qb_pos)
        elif isinstance(gate, tuple(type(gate) for gate in (X, Y, Z, H, S, T))):
            assert get_control_count(cmd) == 0
            gate_str = str(gate).lower()
            qb_pos = cmd.qubits[0][0].id
            self.qasm += "\n{} q[{}]".format(gate_str, qb_pos)
        else:
            raise NotImplementedError('cmd {0} not implemented'.format((cmd,)))

    def _logical_to_physical(self, qb_id: int) -> int:
        """
        Return the physical location of the qubit with the given logical id.

        Args:
            qb_id: ID of the logical qubit whose position should be returned.

        Returns:
            Physical position of logical qubit with id qb_id.
        """
        if self.main_engine.mapper is not None:
            mapping = self.main_engine.mapper.current_mapping
            if qb_id not in mapping:
                raise RuntimeError("Unknown qubit id {}. Please make sure "
                                   "eng.flush() was called and that the qubit "
                                   "was eliminated during optimization."
                                   .format(qb_id))
            return int(mapping[qb_id])
        else:
            return qb_id  # no mapping

    def get_probabilities(self, qureg: List[Qubit]) -> Dict[str, float]:
        """
        Return the list of basis states with corresponding probabilities.

        The measured bits are ordered according to the supplied quantum
        register, i.e., the left-most bit in the state-string corresponds to
        the first qubit in the supplied quantum register.

        Warning:
            Only call this function after the circuit has been executed!

        Args:
            qureg: Quantum register of size n determining the contents of the probability states.

        Returns:
            Dictionary mapping n-bit strings of '0' and '1' to probabilities.

        Raises:
            RuntimeError: If no data is available (i.e., if the circuit has
                not been executed). Or if a qubit was supplied which was not
                present in the circuit (might have gotten optimized away).
        """
        if len(self._measured_states) == 0:
            raise RuntimeError("Please, run the circuit first!")

        mask_bits = map(lambda qubit: self._logical_to_physical(qubit.id), qureg)

        filtered_states = QIBackend._filter_histogram(self._measured_states, mask_bits)

        probability_dict = {self._map_state_to_bit_string(state, qureg): probability
                            for state, probability in filtered_states.items()}

        return probability_dict

    def _map_state_to_bit_string(self, state: int, qureg: List[Qubit]) -> str:
        """

        Args:
            state: state represented as an integer number.
            qureg: list of qubits for which to extract the state bit.

        Returns:
            A string of '0' and '1' corresponding to the bit value in state of each Qubit in qureg.

        Examples:
            state = int('0b101010', 2)
            qureg = [Qubit(0), Qubit(1), Qubit(5)]
            print(self._map_state_to_bit_string(state, qureg)  # prints '011'

        """
        mapped_state = ''

        for i in range(len(qureg)):
            logical_id = qureg[i].id
            physical_id = self._logical_to_physical(logical_id)
            if int(state) & (1 << physical_id):
                mapped_state += '1'
            else:
                mapped_state += '0'

        return mapped_state

    def _run(self) -> None:
        """
        Run the circuit.

        Send the circuit via the Quantum Inspire API.
        """
        if self.qasm == "":
            return

        # Finally: add measurement commands for all measured qubits if no measurements are given.
        # Only measurements after all gate operations will perform properly.
        if not self._measured_ids:
            self.__add_measure_all_qubits()

        self._finalize_qasm()
        self._execute_cqasm()
        self._filter_result_by_measured_qubits()
        self._register_random_measurement_outcome()
        self._reset()

    def _finalize_qasm(self) -> None:
        """ Finalize qasm (add version and qubits line). """
        qasm = 'version 1.0\n# cQASM generated by Quantum Inspire {0} class\nqubits {1}\n\n'.format(
            self.__class__, self._number_of_qubits)
        qasm += self.qasm

        self._cqasm = qasm

    def _execute_cqasm(self) -> None:
        """ Execute self._cqasm through the API.

        Sets self._quantum_inspire_result with the result object in the API response.

        Raises:
            ProjectQBackendError: when raw_text in result from API is not empty (indicating a backend error).
        """
        self._quantum_inspire_result = self.quantum_inspire_api.execute_qasm(
            self._cqasm,
            number_of_shots=self._num_runs,
            backend_type=self.backend_type
        )

        if not self._quantum_inspire_result.get('histogram', {}):
            raw_text = self._quantum_inspire_result.get('raw_text', 'no raw_text in result structure')
            raise ProjectQBackendError(
                'Result structure does not contain proper histogram. raw_text field: %s' % raw_text)

    def _filter_result_by_measured_qubits(self) -> None:
        """ Filters the raw result by collapsing states so that unmeasured qubits are ignored.

        Populates self._measured_states by filtering self._quantum_inspire_result['histogram'] based on
        self._measured_ids (which are supposed to be logical qubit id's).
        """
        mask_bits = map(lambda bit: self._logical_to_physical(bit), self._measured_ids)
        histogram: Dict[int, float] = {int(k): v for k, v in self._quantum_inspire_result['histogram'].items()}
        self._measured_states = QIBackend._filter_histogram(histogram, mask_bits)

    @staticmethod
    def _filter_histogram(histogram: Dict[int, float], mask_bits: Iterator[int]) -> Dict[int, float]:
        """ Filter a histogram by mask_bits.

        Args:
            histogram: input histogram mapping state to probability.
            mask_bits: list of bits that are to be kept in the filtered histogram.

        Returns:
            Collapsed histogram mapping state to probability.

        Keys in a histogram dict are the states represented as an integer number (may be int or string), values the
        probability corresponding to that state.

        The mask_bits list specifies the relevant bits, any bit not set to 1 in the mask will be ignored (masked out).
        The probabilities of equivalent states summed.

        For example, if we have two states b0010 and b0011, and mask_bits is [1] we only care about bit 1 (the second
        bit from the right). The output will specify only one state, b0010, and the probability is the sum of the
        two original probabilities.

        """
        mask = reduce(lambda x, y: x | (1 << y), mask_bits, 0)

        filtered_states: Dict[int, float] = defaultdict(lambda: 0)
        for state, probability in histogram.items():
            filtered_states[state & mask] += probability

        return dict(filtered_states)

    def _register_random_measurement_outcome(self) -> None:
        """ Samples the _measured_states for a single result and registers this as the outcome of the circuit. """

        class QB:
            def __init__(self, qubit_id: int) -> None:
                self.id: int = qubit_id

        random_measurement = self._sample_measured_states_once()

        for logical_qubit_id in self._measured_ids:
            physical_qubit_id = self._logical_to_physical(logical_qubit_id)

            result = bool(random_measurement & (1 << physical_qubit_id))

            self.main_engine.set_measurement_result(QB(logical_qubit_id), result)

    def _sample_measured_states_once(self) -> int:
        """ Obtain a random state from the _measured_states, taking into account the probability distribution. """
        states = list(self._measured_states.keys())
        weights = list(self._measured_states.values())
        return random.choices(states, weights=weights)[0]

    @property
    def _number_of_qubits(self) -> int:
        """ Return the number of qubits used in the circuit. """
        return self._max_qubit_id + 1

    def receive(self, command_list: List[Command]) -> None:
        """
        Receives a command list and, for each command, stores it until completion.

        Args:
            command_list: List of commands to execute.
        """
        for cmd in command_list:
            if not cmd.gate == FlushGate():
                self._store(cmd)
            else:
                self._run()
                self._flushed = True
                self._reset()

    def __add_measure_all_qubits(self) -> None:
        """ Adds measurements at the end of the quantum algorithm for all allocated qubits. """
        qubits_reference = self.main_engine.active_qubits.copy()
        qubits_counts = len(qubits_reference)
        for _ in range(qubits_counts):
            q = qubits_reference.pop()
            Measure | q



"""Example usage of the Quantum Inspire backend with the QisKit SDK.

A simple example that demonstrates how to use the SDK to create
a circuit to create a Bell state, and simulate the circuit on
Quantum Inspire.

For documentation on how to use QisKit we refer to
[https://qiskit.org/](https://qiskit.org/).

Specific to Quantum Inspire is the creation of the QI instance, which is used to set the authentication of the user and
provides a Quantum Inspire backend that is used to execute the circuit.


Copyright 2018-19 QuTech Delft. Licensed under the Apache License, Version 2.0.
"""
import os
from getpass import getpass
from quantuminspire.credentials import load_account, get_token_authentication, get_basic_authentication

from qiskit.validation.base import Obj
from qiskit.circuit import QuantumRegister, ClassicalRegister, QuantumCircuit
from qiskit.tools.compiler import execute

from quantuminspire.qiskit import QI

QI_EMAIL = os.getenv('QI_EMAIL')
QI_PASSWORD = os.getenv('QI_PASSWORD')
QI_URL = os.getenv('API_URL', 'https://api.quantum-inspire.com/')


def get_authentication():
    """ Gets the authentication for connecting to the Quantum Inspire API."""
    token = load_account()
    if token is not None:
        return get_token_authentication(token)
    else:
        if QI_EMAIL is None or QI_PASSWORD is None:
            print('Enter email:')
            email = input()
            print('Enter password')
            password = getpass()
        else:
            email, password = QI_EMAIL, QI_PASSWORD
        return get_basic_authentication(email, password)


if __name__ == '__main__':

    authentication = get_authentication()
    QI.set_authentication(authentication, QI_URL)
    qi_backend = QI.get_backend('QX single-node simulator')

    q = QuantumRegister(2)
    b = ClassicalRegister(2)
    circuit = QuantumCircuit(q, b)

    circuit.h(q[0])
    circuit.cx(q[0], q[1])
    circuit.measure(q, b)

    qi_job = execute(circuit, backend=qi_backend, shots=256)
    qi_result = qi_job.result()
    histogram = qi_result.get_counts(circuit)
    print('\nState\tCounts')
    [print('{0}\t{1}'.format(state, counts)) for state, counts in histogram.items()]
    # Print the full state probabilities histogram
    probabilities_histogram = Obj.to_dict(qi_result.data(circuit)['probabilities'])
    print('\nState\tProbabilities')
    # Format the hexadecimal key to a zero-padded binary string with length of the number of classical bits
    [print('{0}\t{1}'.format(format(int(str(bin(int(key, 16)))[2:], 2), '0{}b'.format(b.size)),
                             val)) for key, val in probabilities_histogram.items()]



"""
This example is copied from https://github.com/ProjectQ-Framework/ProjectQ
and is covered under the Apache 2.0 license.
"""

import os
from getpass import getpass

from projectq import MainEngine
from projectq.backends import ResourceCounter
from projectq.ops import CNOT, H, Measure, All
from projectq.setups import restrictedgateset

from quantuminspire.credentials import load_account, get_token_authentication, get_basic_authentication
from quantuminspire.api import QuantumInspireAPI
from quantuminspire.projectq.backend_qx import QIBackend

QI_EMAIL = os.getenv('QI_EMAIL')
QI_PASSWORD = os.getenv('QI_PASSWORD')
QI_URL = os.getenv('API_URL', 'https://api.quantum-inspire.com/')


def get_authentication():
    """ Gets the authentication for connecting to the Quantum Inspire API."""
    token = load_account()
    if token is not None:
        return get_token_authentication(token)
    else:
        if QI_EMAIL is None or QI_PASSWORD is None:
            print('Enter email:')
            email = input()
            print('Enter password')
            password = getpass()
        else:
            email, password = QI_EMAIL, QI_PASSWORD
        return get_basic_authentication(email, password)


if __name__ == '__main__':

    name = 'TestProjectQ'
    authentication = get_authentication()
    qi_api = QuantumInspireAPI(QI_URL, authentication, project_name=name)

    compiler_engines = restrictedgateset.get_engine_list(one_qubit_gates="any", two_qubit_gates=(CNOT,))
    compiler_engines.extend([ResourceCounter()])

    qi_backend = QIBackend(quantum_inspire_api=qi_api)
    engine = MainEngine(backend=qi_backend, engine_list=compiler_engines)

    qubits = engine.allocate_qureg(2)
    q1 = qubits[0]
    q2 = qubits[1]

    H | q1
    CNOT | (q1, q2)
    All(Measure) | qubits

    engine.flush()

    print('\nMeasured: {0}'.format([int(q) for q in qubits]))
    print('Probabilities {0}'.format(qi_backend.get_probabilities(qubits)))



""" Quantum Inspire SDK

Copyright 2018 QuTech Delft

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import warnings
import os
import json
from typing import Optional
from coreapi.auth import BasicAuthentication, TokenAuthentication

DEFAULT_QIRC_FILE = os.path.join(os.path.expanduser("~"), '.quantuminspire', 'qirc')


def load_account(filename: str = DEFAULT_QIRC_FILE) -> Optional[str]:
    """ Try to load an earlier stored Quantum Inspire token from file or environment

    Load the token when found. This method looks for the token in two locations, in the following order:
    1. In the environment variable ('QI_TOKEN').
    2. In the file with filename given or, when not given, the default resource file in the user home directory
       (`HOME/.quantuminspire/qirc`).

    Args:
        filename: full path to the resource file. If no filename is given, the default resource file
                  in the user home directory is used (`HOME/.quantuminspire/qirc`).

    Returns:
        The Quantum Inspire token or None when no token is found.
    """
    token = os.environ.get('QI_TOKEN', None) or read_account(filename)
    return token


def read_account(filename: str = DEFAULT_QIRC_FILE) -> Optional[str]:
    """ Try to read an earlier stored Quantum Inspire token from file

    Read the token from file. This method looks for the token in the file with filename given or,
    when no filename is given, the default resource file in the user home directory (`HOME/.quantuminspire/qirc`).

    Args:
        filename: full path to the resource file. If no filename is given, the default resource file
                  in the user home directory is used (`HOME/.quantuminspire/qirc`).

    Returns:
        The Quantum Inspire token or None when no token is found.
    """
    try:
        with open(filename, 'r') as file:
            accounts = json.load(file)
            token: Optional[str] = accounts['token']
    except (OSError, KeyError, ValueError):  # file does not exist or is empty/invalid
        token = None
    return token if (token and len(token)) else None


def store_account(token: str, filename: str = DEFAULT_QIRC_FILE, overwrite: bool = False) -> None:
    """
    Store the token in a resource file. Replace an existing token only when overwrite=True.

    Args:
        token: the Quantum Inspire token to store to disk.
        filename: full path to the resource file. If no filename is given, the default resource file
                  in the user home directory is used (`HOME/.quantuminspire/qirc`).
        overwrite: overwrite an existing token.
    """
    stored_token = read_account(filename)
    if stored_token and stored_token != token and not overwrite:
        warnings.warn('Token already present. Set overwrite=True to overwrite.')
        return
    save_account(token, filename)


def delete_account(token: str, filename: str = DEFAULT_QIRC_FILE) -> None:
    """ Remove the token from the resource file.

    Args:
        token: the Quantum Inspire token to remove.
        filename: full path to the resource file. If no filename is given, the default resource file
                  in the user home directory is used (`HOME/.quantuminspire/qirc`).
    """
    stored_token = read_account(filename)
    if stored_token == token:
        save_account('', filename)


def save_account(token: str, filename: str = DEFAULT_QIRC_FILE) -> None:
    """ Save the token to a file with filename given, otherwise save to the default resource file.
        An existing token is overwritten. Use store_account to prevent overwriting an existing token.

    Args:
        token: the Quantum Inspire token to save.
        filename: full path to the resource file. If no filename is given, the default resource file
                  in the user home directory is used (`HOME/.quantuminspire/qirc`).
    """
    accounts = {'token': token}
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, 'w') as config_file:
        json.dump(accounts, config_file, indent=2)


def enable_account(token: str) -> None:
    """ Save the token to the internal environment, that will be used by load_account for the session.
        When a token was already loaded from the system environment it is overwritten.
        The system environment is not effected.

    Args:
        token: the Quantum Inspire token to be used by load_account() for the session.
    """
    os.environ['QI_TOKEN'] = token


def get_token_authentication(token: Optional[str] = None) -> TokenAuthentication:
    """ Set up token authentication for Quantum Inspire to be used in the API.

    Args:
        token: the Quantum Inspire token to set in TokenAuthentication. When no token is given,

    Returns:
        The token authentication for Quantum Inspire.
    """
    if not token:
        token = load_account()
    return TokenAuthentication(token, scheme="token")


def get_basic_authentication(email: str, password: str) -> BasicAuthentication:
    """ Set up basic authentication for Quantum Inspire to be used in the API.

    Args:
        email: a valid email address.
        password: password for the account.

    Returns:
        The basic authentication for Quantum Inspire.
    """
    return BasicAuthentication(email, password)



""" Quantum Inspire SDK

Copyright 2018 QuTech Delft

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""


class QisKitBackendError(Exception):
    """ Exception for SDK errors related to the qiskit backend."""


class ProjectQBackendError(Exception):
    """ Exception for SDK errors related to the projectq backend."""


class AuthenticationError(Exception):
    """ Exception for SDK errors related to authentication."""


class ApiError(Exception):
    """ Exception for SDK errors related to the API functionality."""



from copy import copy
from typing import List, Optional, Any

import coreapi
from qiskit.providers import BaseProvider

from quantuminspire.credentials import get_token_authentication, get_basic_authentication
from quantuminspire.qiskit.backend_qx import QuantumInspireBackend
from quantuminspire.api import QuantumInspireAPI
from quantuminspire.exceptions import ApiError

QI_URL = 'https://api.quantum-inspire.com'


class QuantumInspireProvider(BaseProvider):  # type: ignore
    """ Provides a backend and an api for a single Quantum Inspire account. """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._backends: List[QuantumInspireBackend] = []
        self._api: Optional[QuantumInspireAPI] = None

    def __str__(self) -> str:
        return 'QI'

    def backends(self, name: Optional[str] = None, **kwargs: Any) -> List[QuantumInspireBackend]:
        """
        Provides a list of backends.

        Args:
            name: Name of the requested backend.
            **kwargs: Used for filtering, not implemented.

        Returns:
            List of backends that meet the filter requirements.
        """
        if self._api is None:
            raise ApiError('Authentication details have not been set.')

        available_backends = self._api.get_backend_types()
        if name is not None:
            available_backends = list(filter(lambda b: b['name'] == name, available_backends))
        backends = []
        for backend in available_backends:
            if backend['is_allowed']:
                config = copy(QuantumInspireBackend.DEFAULT_CONFIGURATION)
                config.backend_name = backend['name']
                backends.append(QuantumInspireBackend(self._api, provider=self, configuration=config))

        return backends

    def set_authentication_details(self, email: str, password: str, qi_url: str = QI_URL) -> None:
        """
        DEPRECATED(version>'0.5.0', reason="Replaced with method set_basic_authentication(email, password, qi_url)")
        Set a single authentication for Quantum Inspire.

        Args:
            email: A valid email address.
            password: Password for the account.
            qi_url: URL that points to quantum-inspire api. Default value: 'https://api.quantum-inspire.com'.

        """
        self.set_basic_authentication(email, password)

    def set_basic_authentication(self, email: str, password: str, qi_url: str = QI_URL) -> None:
        """
        Set up basic authentication for Quantum Inspire.

        Args:
            email: A valid email address.
            password: Password for the account.
            qi_url: URL that points to quantum-inspire api. Default value: 'https://api.quantum-inspire.com'.
        """
        authentication = get_basic_authentication(email, password)
        self.set_authentication(authentication, qi_url)

    def set_token_authentication(self, token: str, qi_url: str = QI_URL) -> None:
        """
        Set up token authentication for Quantum Inspire.

        Args:
            token: A valid token.
            qi_url: URL that points to quantum-inspire api. Default value: 'https://api.quantum-inspire.com'.
        """
        authentication = get_token_authentication(token)
        self.set_authentication(authentication, qi_url)

    def set_authentication(self, authentication: Optional[coreapi.auth.AuthBase] = None,
                           qi_url: str = QI_URL) -> None:
        """
        Initializes the API and sets the authentication for Quantum Inspire.

        Args:
            authentication: The authentication, can be one of the following coreapi authentications:
                            BasicAuthentication(email, password), HTTP authentication with valid email/password.
                            TokenAuthentication(token, scheme="token"), token authentication with a valid API-token.
                            When authentication is None, api will try to load a token from the default resource.
            qi_url: URL that points to quantum-inspire api. Default value: 'https://api.quantum-inspire.com'.
        """
        self._api = QuantumInspireAPI(qi_url, authentication)



"""
This example is copied from https://github.com/ProjectQ-Framework/ProjectQ
and is covered under the Apache 2.0 license.
"""
import os
import math
from getpass import getpass

from projectq import MainEngine
from projectq.backends import ResourceCounter, Simulator
from projectq.meta import Compute, Control, Loop, Uncompute
from projectq.ops import CNOT, CZ, All, H, Measure, Toffoli, X, Z
from projectq.setups import restrictedgateset

from quantuminspire.credentials import load_account, get_token_authentication, get_basic_authentication
from quantuminspire.api import QuantumInspireAPI
from quantuminspire.projectq.backend_qx import QIBackend

QI_EMAIL = os.getenv('QI_EMAIL')
QI_PASSWORD = os.getenv('QI_PASSWORD')
QI_URL = os.getenv('API_URL', 'https://api.quantum-inspire.com/')


def run_grover(eng, n, oracle):
    """
    Runs Grover's algorithm on n qubit using the provided quantum oracle.

    Args:
        eng (MainEngine): Main compiler engine to run Grover on.
        n (int): Number of bits in the solution.
        oracle (function): Function accepting the engine, an n-qubit register,
            and an output qubit which is flipped by the oracle for the correct
            bit string.

    Returns:
        solution (list<int>): Solution bit-string.
    """
    x = eng.allocate_qureg(n)

    # start in uniform superposition
    All(H) | x

    # number of iterations we have to run:
    num_it = int(math.pi / 4. * math.sqrt(1 << n))

    # prepare the oracle output qubit (the one that is flipped to indicate the
    # solution. start in state 1/sqrt(2) * (|0> - |1>) s.t. a bit-flip turns
    # into a (-1)-phase.
    oracle_out = eng.allocate_qubit()
    X | oracle_out
    H | oracle_out

    # run num_it iterations
    with Loop(eng, num_it):
        # oracle adds a (-1)-phase to the solution
        oracle(eng, x, oracle_out)

        # reflection across uniform superposition
        with Compute(eng):
            All(H) | x
            All(X) | x

        with Control(eng, x[0:-1]):
            Z | x[-1]

        Uncompute(eng)

    All(Measure) | x
    Measure | oracle_out

    eng.flush()
    # return result
    return [int(qubit) for qubit in x]


def alternating_bits_oracle(eng, qubits, output):
    """
    Marks the solution string 1,0,1,0,...,0,1 by flipping the output qubit,
    conditioned on qubits being equal to the alternating bit-string.

    Args:
        eng (MainEngine): Main compiler engine the algorithm is being run on.
        qubits (Qureg): n-qubit quantum register Grover search is run on.
        output (Qubit): Output qubit to flip in order to mark the solution.
    """
    with Compute(eng):
        All(X) | qubits[1::2]
    with Control(eng, qubits):
        X | output
    Uncompute(eng)


def get_authentication():
    """ Gets the authentication for connecting to the Quantum Inspire API."""
    token = load_account()
    if token is not None:
        return get_token_authentication(token)
    else:
        if QI_EMAIL is None or QI_PASSWORD is None:
            print('Enter email:')
            email = input()
            print('Enter password')
            password = getpass()
        else:
            email, password = QI_EMAIL, QI_PASSWORD
        return get_basic_authentication(email, password)


if __name__ == '__main__':

    # Remote Quantum-Inspire backend #
    authentication = get_authentication()
    qi = QuantumInspireAPI(QI_URL, authentication)

    compiler_engines = restrictedgateset.get_engine_list(one_qubit_gates="any", two_qubit_gates=(CNOT, CZ, Toffoli))
    compiler_engines.extend([ResourceCounter()])
    qi_backend = QIBackend(quantum_inspire_api=qi)
    qi_engine = MainEngine(backend=qi_backend, engine_list=compiler_engines)

    # Run remote Grover search to find a n-bit solution
    result_qi = run_grover(qi_engine, 3, alternating_bits_oracle)
    print("\nResult from the remote Quantum-Inspire backend: {}".format(result_qi))

    # Local ProjectQ simulator backend #
    compiler_engines = restrictedgateset.get_engine_list(one_qubit_gates="any", two_qubit_gates=(CNOT, CZ))
    compiler_engines.append(ResourceCounter())
    local_engine = MainEngine(Simulator(), compiler_engines)

    # Run local Grover search to find a n-bit solution
    result_local = run_grover(local_engine, 3, alternating_bits_oracle)
    print("Result from the local ProjectQ simulator backend: {}\n".format(result_local))



""" Quantum Inspire SDK

Copyright 2018 QuTech Delft

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""
import copy
import numpy as np
from io import StringIO
from typing import Optional
from qiskit.qobj import QobjItem
from quantuminspire.exceptions import ApiError


class CircuitToString:
    """ Contains the translational elements to convert the Qiskit circuits to cQASM code."""

    @staticmethod
    def _gate_not_supported(_stream: StringIO, instruction: QobjItem, _binary_control: Optional[str] = None) -> None:
        """ Called when a gate is not supported with the backend. Throws an exception (ApiError)

        Args:
            instruction: The Qiskit instruction to translate to cQASM.

        Raises:
            ApiError: the gate is not supported by the circuit parser.

        """
        if hasattr(instruction, 'conditional'):
            raise ApiError('Conditional gate c-{} not supported'.format(instruction.name.lower()))
        else:
            raise ApiError('Gate {} not supported'.format(instruction.name.lower()))

    @staticmethod
    def _cx(stream: StringIO, instruction: QobjItem) -> None:
        """ Translates the controlled X element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('CNOT q[{0}], q[{1}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_cx(stream: StringIO, instruction: QobjItem, binary_control: str) -> None:
        """ Translates the binary-controlled controlled X element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.
            binary_control: The multi-bits control string. The gate is executed when all specified classical bits are 1.

        """
        stream.write('C-CNOT {0}q[{1}], q[{2}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _ccx(stream: StringIO, instruction: QobjItem) -> None:
        """ Translates the Toffoli element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('Toffoli q[{0}], q[{1}], q[{2}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_ccx(stream: StringIO, instruction: QobjItem, binary_control: str) -> None:
        """ Translates the binary controlled Toffoli element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.
            binary_control: The multi-bits control string. The gate is executed when all specified classical bits are 1.

        """
        stream.write('C-Toffoli {0}q[{1}], q[{2}], q[{3}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _h(stream: StringIO, instruction: QobjItem) -> None:
        """ Translates the H element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('H q[{0}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_h(stream: StringIO, instruction: QobjItem, binary_control: str) -> None:
        """ Translates the binary-controlled H element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.
            binary_control: The multi-bits control string. The gate is executed when all specified classical bits are 1.

        """
        stream.write('C-H {0}q[{1}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _id(stream: StringIO, instruction: QobjItem) -> None:
        """ Translates the ID element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('I q[{0}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_id(stream: StringIO, instruction: QobjItem, binary_control: str) -> None:
        """ Translates the binary-controlled ID element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.
            binary_control: The multi-bits control string. The gate is executed when all specified classical bits are 1.

        """
        stream.write('C-I {0}q[{1}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _s(stream: StringIO, instruction: QobjItem) -> None:
        """ Translates the S element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('S q[{0}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_s(stream: StringIO, instruction: QobjItem, binary_control: str) -> None:
        """ Translates the binary-controlled S element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('C-S {0}q[{1}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _sdg(stream: StringIO, instruction: QobjItem) -> None:
        """ Translates the Sdag element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('Sdag q[{0}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_sdg(stream: StringIO, instruction: QobjItem, binary_control: str) -> None:
        """ Translates the binary-controlled Sdag element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('C-Sdag {0}q[{1}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _swap(stream: StringIO, instruction: QobjItem) -> None:
        """ Translates the SWAP element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('SWAP q[{0}], q[{1}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_swap(stream: StringIO, instruction: QobjItem, binary_control: str) -> None:
        """ Translates the binary-controlled SWAP element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('C-SWAP {0}q[{1}], q[{2}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _t(stream: StringIO, instruction: QobjItem) -> None:
        """ Translates the T element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('T q[{0}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_t(stream: StringIO, instruction: QobjItem, binary_control: str) -> None:
        """ Translates the binary-controlled T element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('C-T {0}q[{1}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _tdg(stream: StringIO, instruction: QobjItem) -> None:
        """ Translates the Tdag element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('Tdag q[{0}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_tdg(stream: StringIO, instruction: QobjItem, binary_control: str) -> None:
        """ Translates the binary-controlled Tdag element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('C-Tdag {0}q[{1}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _x(stream: StringIO, instruction: QobjItem) -> None:
        """ Translates the X element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('X q[{0}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_x(stream: StringIO, instruction: QobjItem, binary_control: str) -> None:
        """ Translates the binary-controlled X element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('C-X {0}q[{1}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _y(stream: StringIO, instruction: QobjItem) -> None:
        """ Translates the Y element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('Y q[{0}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_y(stream: StringIO, instruction: QobjItem, binary_control: str) -> None:
        """ Translates the binary-controlled Y element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.
            binary_control: The multi-bits control string. The gate is executed when all specified classical bits are 1.

        """
        stream.write('C-Y {0}q[{1}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _z(stream: StringIO, instruction: QobjItem) -> None:
        """ Translates the Z element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        stream.write('Z q[{0}]\n'.format(*instruction.qubits))

    @staticmethod
    def _c_z(stream: StringIO, instruction: QobjItem, binary_control: str) -> None:
        """ Translates the binary-controlled Z element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.
            binary_control: The multi-bits control string. The gate is executed when all specified classical bits are 1.

        """
        stream.write('C-Z {0}q[{1}]\n'.format(binary_control, *instruction.qubits))

    @staticmethod
    def _r(stream: StringIO, instruction: QobjItem, axis: str) -> None:
        """ Translates the Rotation element for an axis (x,y,z).

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.
            axis: The axis for which the Rotation operator is parsed ('x', 'y' or 'z').

        """
        angle_q0 = instruction.params[0]
        stream.write('R{0} q[{1}], {2:.6f}\n'.format(axis, *instruction.qubits, angle_q0))

    @staticmethod
    def _c_r(stream: StringIO, instruction: QobjItem, axis: str, binary_control: str) -> None:
        """ Translates the binary-controlled Rotation element for an axis (x,y,z).

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.
            axis: The axis for which the Rotation operator is parsed ('x', 'y' or 'z').
            binary_control: The multi-bits control string. The gate is executed when all specified classical bits are 1.

        """
        angle_q0 = instruction.params[0]
        stream.write('C-R{0} {1}q[{2}], {3:.6f}\n'.format(axis, binary_control, *instruction.qubits, angle_q0))

    @staticmethod
    def _rx(stream: StringIO, instruction: QobjItem) -> None:
        """ Translates the Rx element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        CircuitToString._r(stream, instruction, 'x')

    @staticmethod
    def _c_rx(stream: StringIO, instruction: QobjItem, binary_control: str) -> None:
        """ Translates the binary-controlled Rx element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.
            binary_control: The multi-bits control string. The gate is executed when all specified classical bits are 1.

        """
        CircuitToString._c_r(stream, instruction, 'x', binary_control)

    @staticmethod
    def _ry(stream: StringIO, instruction: QobjItem) -> None:
        """ Translates the Ry element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        CircuitToString._r(stream, instruction, 'y')

    @staticmethod
    def _c_ry(stream: StringIO, instruction: QobjItem, binary_control: str) -> None:
        """ Translates the binary-controlled Ry element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.
            binary_control: The multi-bits control string. The gate is executed when all specified classical bits are 1.

        """
        CircuitToString._c_r(stream, instruction, 'y', binary_control)

    @staticmethod
    def _rz(stream: StringIO, instruction: QobjItem) -> None:
        """ Translates the Rz element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        CircuitToString._r(stream, instruction, 'z')

    @staticmethod
    def _c_rz(stream: StringIO, instruction: QobjItem, binary_control: str) -> None:
        """ Translates the binary-controlled Rz element.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.
            binary_control: The multi-bits control string. The gate is executed when all specified classical bits are 1.

        """
        CircuitToString._c_r(stream, instruction, 'z', binary_control)

    @staticmethod
    def _u(stream: StringIO, instruction: QobjItem) -> None:
        """ Translates the U element to U3. The u element is used by qiskit for the u_base gate and when a u0-gate
            is used in the circuit but not supported as a basis gate for the backend.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        CircuitToString._u3(stream, instruction)

    @staticmethod
    def _c_u(stream: StringIO, instruction: QobjItem, binary_control: str) -> None:
        """ Translates the binary-controlled U element to binary-controlled U3.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.
            binary_control: The multi-bits control string. The gate is executed when all specified classical bits are 1.

        """
        CircuitToString._c_u3(stream, instruction, binary_control)

    @staticmethod
    def _u1(stream: StringIO, instruction: QobjItem) -> None:
        """ Translates the U1(lambda) element to U3(0, 0, lambda). A copy of the circuit is made to prevent
            side-effects for the caller.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        temp_instruction = copy.deepcopy(instruction)
        temp_instruction.params[0:0] = (0, 0)
        temp_instruction.texparams[0:0] = ('0', '0')
        CircuitToString._u3(stream, temp_instruction)

    @staticmethod
    def _c_u1(stream: StringIO, instruction: QobjItem, binary_control: str) -> None:
        """ Translates the binary-controlled U1(lambda) element to U3(0, 0, lambda). A copy of the circuit is
        made to prevent side-effects for the caller.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.
            binary_control: The multi-bits control string. The gate is executed when all specified classical bits are 1.

        """
        temp_instruction = copy.deepcopy(instruction)
        temp_instruction.params[0:0] = (0, 0)
        temp_instruction.texparams[0:0] = ('0', '0')
        CircuitToString._c_u3(stream, temp_instruction, binary_control)

    @staticmethod
    def _u2(stream: StringIO, instruction: QobjItem) -> None:
        """ Translates the U2(phi, lambda) element to U3(pi/2, phi, lambda). A copy of the circuit is made to prevent
            side-effects for the caller.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        temp_instruction = copy.deepcopy(instruction)
        temp_instruction.params.insert(0, np.pi/2)
        temp_instruction.texparams.insert(0, '\\frac{\\pi}{2}')
        CircuitToString._u3(stream, temp_instruction)

    @staticmethod
    def _c_u2(stream: StringIO, instruction: QobjItem, binary_control: str) -> None:
        """ Translates the binary-controlled U2(phi, lambda) element to U3(pi/2, phi, lambda). A copy of the
        circuit is made to prevent side-effects for the caller.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.
            binary_control: The multi-bits control string. The gate is executed when all specified classical bits are 1.

        """
        temp_instruction = copy.deepcopy(instruction)
        temp_instruction.params.insert(0, np.pi/2)
        temp_instruction.texparams.insert(0, '\\frac{\\pi}{2}')
        CircuitToString._c_u3(stream, temp_instruction, binary_control)

    @staticmethod
    def _u3(stream: StringIO, instruction: QobjItem) -> None:
        """ Translates the U3(theta, phi, lambda) element to 3 rotation gates.
            Any single qubit operation (a 2x2 unitary matrix) can be written as the product of rotations.
            As an example, a unitary single-qubit gate can be expressed as a combination of
            Rz and Ry rotations (Nielsen and Chuang, 10th edition, section 4.2).
            U(theta, phi, lambda) = Rz(phi)Ry(theta)Rz(lambda).
            Note: The expression above is the matrix multiplication, when implementing this in a gate circuit,
            the gates need to be executed in reversed order.
            Any rotation of 0 radials is left out of the resulting circuit.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        gates = ['Rz', 'Ry', 'Rz']
        angles = list(instruction.params[i] for i in [2, 0, 1])
        index_q0 = [instruction.qubits[0]] * 3
        for triplet in zip(gates, index_q0, angles):
            if triplet[2] != 0:
                stream.write('{0} q[{1}], {2:.6f}\n'.format(*triplet))

    @staticmethod
    def _c_u3(stream: StringIO, instruction: QobjItem, binary_control: str) -> None:
        """ Translates the binary-controlled U3(theta, phi, lambda) element to 3 rotation gates.
            See gate _u3 for more information.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.
            binary_control: The multi-bits control string. The gate is executed when all specified classical bits are 1.

        """
        gates = ['C-Rz', 'C-Ry', 'C-Rz']
        binary_controls = [binary_control] * 3
        angles = list(instruction.params[i] for i in [2, 0, 1])
        index_q0 = [instruction.qubits[0]] * 3
        for quadruplets in zip(gates, binary_controls, index_q0, angles):
            if quadruplets[3] != 0:
                stream.write('{0} {1}q[{2}], {3:.6f}\n'.format(*quadruplets))

    @staticmethod
    def _barrier(stream: StringIO, instruction: QobjItem) -> None:
        """ Translates the | element. No cQASM is added for this gate.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        pass

    @staticmethod
    def _c_barrier(stream: StringIO, instruction: QobjItem, binary_control: str) -> None:
        """ Translates the binary-controlled | element. No cQASM is added for this gate.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.
            binary_control: The multi-bits control string. The gate is executed when all specified classical bits are 1.

        """
        pass

    @staticmethod
    def _measure(stream: StringIO, instruction: QobjItem) -> None:
        """ Translates the measure element. No cQASM is added for this gate.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        pass

    def _parse_bin_ctrl_gate(self, stream: StringIO, instruction: QobjItem) -> None:
        """ Parses a binary controlled gate. A binary controlled gate name is preceded by 'c-'.
            The gate is executed when a specific measurement is true. Multiple measurement outcomes are used
            to control the quantum operation. This measurement is a combination of classical bits being 1 and others
            being 0. Because cQASM only supports measurement outcomes of 1, any other bits in the
            masked bit pattern first have to be inverted with the not-operator. The same inversion also has to
            take place after the binary controlled quantum operation.
            The resulting stream will be expanded with something like:
            not b[the 0-bits changed to 1]
            c-gate [classical bits], other arguments
            not b[the 0-bits reset to 0 again]
            When the c-gate results in an empty string (e.g. binary controlled u(0, 0, 0) or barrier gate),
            nothing is added to the stream.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.

        """
        conditional_type = instruction.conditional.type
        if conditional_type != 'equals':
            raise ApiError('Conditional statement with type {} not supported'.format(conditional_type))
        mask = int(instruction.conditional.mask, 16)
        if mask == 0:
            raise ApiError('Conditional statement {} without a mask'.format(instruction.name.lower()))
        nr_of_classical_bits = int(np.log2(mask + 1))
        val = int(instruction.conditional.val, 16)
        masked_val = mask & val

        # form the negation to the 0-values of the measurement registers, when mask == val no bits are negated
        negate_zeroes_line = ''
        if masked_val != mask:
            negate_zeroes_line = 'not b[' + ','.join(
                str(i) for i in range(nr_of_classical_bits) if not (masked_val & (1 << i))) + ']\n'

        # form multi bits control - qasm-single-gate-multiple-qubits
        binary_control = 'b[0:{}], '.format(nr_of_classical_bits - 1)

        with StringIO() as gate_stream:
            # add the gate
            gate_name = '_c_%s' % instruction.name.lower()
            gate_function = getattr(self, gate_name, getattr(self, "_gate_not_supported"))
            gate_function(gate_stream, instruction, binary_control)
            line = gate_stream.getvalue()
            if len(line) != 0:
                # negate the measurement registers that has to be 0
                stream.write(negate_zeroes_line)
                stream.write(line)
                # reverse the measurement registers that had to be 0
                stream.write(negate_zeroes_line)

    def parse(self, stream: StringIO, instruction: QobjItem) -> None:
        """ Parses a gate. For each type of gate a separate (private) parsing method is defined and called.
            The resulting cQASM code is written to the stream. When the gate is a binary controlled
            gate, the parsing is forwarded to method _parse_bin_ctrl_gate. When a gate is not supported
            _gate_not_supported is called which raises an exception.

        Args:
            stream: The string-io stream to where the resulting cQASM is written.
            instruction: The Qiskit instruction to translate to cQASM.
        """
        if hasattr(instruction, 'conditional'):
            self._parse_bin_ctrl_gate(stream, instruction)
        else:
            gate_name = '_%s' % instruction.name.lower()
            gate_function = getattr(self, gate_name, getattr(self, "_gate_not_supported"))
            gate_function(stream, instruction)



import time
from typing import List, Optional, Any

from qiskit.providers import BaseJob, JobStatus, JobError, JobTimeoutError
from qiskit.qobj import Qobj, QobjExperiment
from qiskit.result import Result
from quantuminspire import __version__ as quantum_inspire_version
from quantuminspire.api import QuantumInspireAPI


class QIJob(BaseJob):  # type: ignore
    """
    A job that is to be executed on the Quantum-inspire platform. A QIJob is normally created by calling run on the
    QuantumInspireBackend but can also be recreated using a job_id:

            qi_backend = QI.get_backend('QX single-node simulator')
            job = qi_backend.retrieve_job(job_id)
            result = job.result()
    """

    def __init__(self, backend: Any, job_id: str, api: QuantumInspireAPI, qobj: Optional[Qobj] = None) -> None:
        """
        Construct a new QIJob object. Not normally called directly, use a backend object to create/retrieve jobs.

        Args:
            backend: A quantum-inspire backend.
            job_id: Id of the job as provided by the quantum-inspire api.
            api: A quantum-inspire api.
            qobj: A qiskit quantum object.
        """
        self._api: QuantumInspireAPI = api
        super().__init__(backend, job_id)
        self.experiments: Optional[List[QobjExperiment]] = None
        self._status: str = JobStatus.INITIALIZING
        self._qobj: Optional[Qobj] = qobj
        if self._qobj is not None:
            self._job_id = ''  # invalidate _job_id
        else:
            self.status()

    def submit(self) -> None:
        """
        Submit a job to the quantum-inspire platform.

        Raises:
             JobError: An error if the job has already been submitted.
        """
        if self._job_id:
            raise JobError('Job has already been submitted!')
        self._job_id = self._backend.run(self._qobj)

    def result(self, timeout: Optional[float] = None, wait: float = 0.5) -> Result:
        """

        Args:
            timeout: Timeout in seconds.
            wait: Wait time between queries to the quantum-inspire platform.

        Returns:
            Result object containing results from the experiments.

        Raises:
            JobTimeoutError: If timeout is reached.
            QisKitBackendError: If an error occurs during simulation.
        """
        start_time = time.time()
        while self.status() != JobStatus.DONE:
            elapsed_time = time.time() - start_time
            if timeout is not None and elapsed_time > timeout:
                raise JobTimeoutError('Failed getting result: timeout reached.')
            time.sleep(wait)
        experiment_results = self._backend.get_experiment_results(self)
        return Result(backend_name=self._backend.backend_name, backend_version=quantum_inspire_version,
                      job_id=self.job_id(), qobj_id=self.job_id(), success=True, results=experiment_results)

    def cancel(self) -> None:
        """ Cancel the job and delete the project. """
        self._api.delete_project(int(self._job_id))

    def status(self) -> JobStatus:
        """
        Query the quantum-inspire platform for the status of the job.

        Returns:
            The status of the job.
        """
        jobs = self._api.get_jobs_from_project(int(self._job_id))
        number_of_jobs = len(jobs)
        cancelled = len([job for job in jobs if job['status'] == 'CANCELLED'])
        running = len([job for job in jobs if job['status'] == 'RUNNING'])
        completed = len([job for job in jobs if job['status'] == 'COMPLETE'])

        if 0 < cancelled < number_of_jobs:
            self._status = JobStatus.ERROR
        elif cancelled == number_of_jobs:
            self._status = JobStatus.CANCELLED
        elif running > 0 or (0 < completed < number_of_jobs):
            self._status = JobStatus.RUNNING
        elif completed == number_of_jobs:
            self._status = JobStatus.DONE
        else:
            self._status = JobStatus.QUEUED
        return self._status



""" Quantum Inspire SDK

Copyright 2018 QuTech Delft

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import itertools
import time
import uuid
from typing import Type, List, Dict, Union, Optional, Any
from collections import OrderedDict
from urllib.parse import urljoin
import coreapi
from coreapi.auth import TokenAuthentication
from coreapi.exceptions import CoreAPIException, ErrorMessage

from quantuminspire.credentials import load_account
from quantuminspire.exceptions import ApiError, AuthenticationError
from quantuminspire.job import QuantumInspireJob

QI_URL = 'https://api.quantum-inspire.com'


class QuantumInspireAPI:

    def __init__(self, base_uri: str = QI_URL, authentication: Optional[coreapi.auth.AuthBase] = None,
                 project_name: Optional[str] = None,
                 coreapi_client_class: Type[coreapi.Client] = coreapi.Client) -> None:
        """ Python interface to the Quantum Inspire API (Application Programmer Interface).

        The Quantum Inspire API supplies an interface for executing cQASM programs and can be used to access the
        different entities in the Quantum Inspire database needed for running the programs.
        The entities for which an interface is provided are:
            Backend types: Depending on the user account more qubits can be used and simulated on faster hardware.
            Projects: Executing programs is done from a project. Projects keep track of the other entities.
            Jobs: A job contains all the information and the parameters needed for executing the program.
            Assets: A container for a cQASM program. Is part of a job.
            Results: After the job is finished, the results are gathered in the result-entity.

        QuantumInspireAPI is a convenient interface (or wrapper) to the low level API and hides details for
        requesting data (via get) and performing operations on the different entities (via actions).

        For more documentation see the knowledge base on: https://www.quantum-inspire.com/kbase/low-level-api/
        The REST API can be found on: https://api.quantum-inspire.com/
        The Core API schema is published on: https://api.quantum-inspire.com/schema/

        Args:
            base_uri: The base uri of the Quantum Inspire API-location where the schema can be found (in path
                      'schema/').
            authentication: The authentication, can be one of the following coreapi authentications:
                            BasicAuthentication(email, password), HTTP authentication with valid email/password.
                            TokenAuthentication(token, scheme="token"), token authentication with a valid API-token.
                            When authentication is None, a token is read from the default location.
            project_name: The project used for executing the jobs.
            coreapi_client_class: Coreapi client to interact with the API through a schema.
                                  Default set to coreapi.Client.

        Note: When no project name is given, a temporary project is created for the job and deleted after the job
              has finished. When a project name is given, a project is created if it does not exist, but re-used
              if a project with that name already exists. In either case, the project will not be deleted when a
              project name is supplied here.

        Raises:
            AuthenticationError: An AuthenticationError exception is raised when no authentication is given
                                 and the token could not be loaded from the default location.
            ApiError: An ApiError exception is raised when the schema could not be loaded.
        """
        if authentication is None:
            token = load_account()
            if token is not None:
                authentication = TokenAuthentication(token, scheme="token")
            else:
                raise AuthenticationError('No credentials have been provided or found on disk')
        self.__client = coreapi_client_class(auth=authentication)
        self.project_name = project_name
        self.base_uri = base_uri
        try:
            self._load_schema()
        except (CoreAPIException, TypeError) as ex:
            raise ApiError('Could not connect to {}'.format(base_uri)) from ex

    def _get(self, uri_path: str) -> Any:
        """ Method for making requests to the coreapi client instance to get some piece of information. The information
            requested depends on the uri_path parameter.

        Args:
            uri_path: The URL where to request the data.

        Note:
            This method will be made private in the near future. Usage is discouraged.

        Raises:
            TypeError when the uri_path is not correct.
            A CoreAPIException is thrown when the get was not successful. Possible causes are:
            the schema is not loaded or a network error occurred.
            A specific CoreAPIException, ErrorMessage, is raised when the result of the get is not successful.

        Returns:
            The resulting data from the get-request. The structure of the data depends on the request.
        """
        return self.__client.get(uri_path)

    def _action(self, action: List[str], params: Optional[Dict[str, Any]] = None) -> Any:
        """ Adapter for performing an action on an object via the Quantum Inspire API.

        Args:
            action: Path in the schema hierarchy selecting the requested action.
            params: Some actions may accept a set of parameters with names as keys.

        Raises:
            A CoreAPIException is thrown when the action was not successful. Possible causes are:
            One of the actions or parameters not supported, the schema is not loaded or a network error occurred.
            A specific CoreAPIException, ErrorMessage, is raised when the result of the action is not successful.

        Returns:
            The resulting data from the action-request. The structure of the data depends on the request.
            Can be None when there is no content in the respons.
        """
        return self.__client.action(self.document, action, params=params)

    def _load_schema(self) -> None:
        """ Loads the schema with metadata that explains how the api-data is structured."""
        self.document = self._get(urljoin(self.base_uri, 'schema/'))

    def list_backend_types(self) -> None:
        """ Prints the backend types with the name and the maximum number of qubits it supports."""
        backends = self.get_backend_types()
        for backend in backends:
            print(f'Backend type: {backend["name"]}, number of qubits: {backend["number_of_qubits"]}')

    def get_default_backend_type(self) -> Dict[str, Any]:
        """ Gets the properties of the default backend type.

        Returns:
            The default backend type with all of its properties:
                | key                        | description
                |----------------------------|-------------------------------------------------------------------------
                | url (str)                  | The url for the backend type.
                | name (str)                 | Name of the backend.
                | is_hardware_backend (bool) | Indicates whether the backend is a hardware backend (True) or
                |                            | a simulating backend (False).
                | required_permission (str)  | Describes the permission that is required to use this backend.
                | number_of_qubits (int)     | Maximum number of qubits the backend supports.
                | description (str)          | Short description of the backend.
                | topology (OrderedDict)     | Dictionary with property 'edges' (list), specifies a list of tuples
                |                            | that define qubit connectivity for 2-qubit gates.
                | is_allowed (bool)          | Indicates whether the user is allowed to use this backend.
        """
        return OrderedDict(self._action(['backendtypes', 'default', 'list']))

    def get_backend_types(self) -> List[Dict[str, Any]]:
        """ Gets a list of backend types with properties.

        Returns:
            Returns a list of backend types with all of its properties.
            See `get_default_backend_type` for a description of the backend properties.
        """
        ret: List[Dict[str, Any]] = self._action(['backendtypes', 'list'])
        return ret

    def get_backend_type_by_id(self, backend_type_id: int) -> Dict[str, Any]:
        """ Gets the properties of a specific backend type, given the backend type id.

        Args:
            backend_type_id: The backend identification number.

        Raises:
            ApiError: An ApiError exception is raised when the backend type indicated by backend_type_id does not exist.

        Returns:
            The requested backend type indicated by backend_type_id with all of its properties.
            See `get_default_backend_type` for a description of the backend type properties.
        """
        try:
            backend_type = self._action(['backendtypes', 'read'], params={'id': backend_type_id})
        except ErrorMessage as err_msg:
            raise ApiError('Backend type with id {} does not exist!'.format(backend_type_id)) from err_msg
        return OrderedDict(backend_type)

    def get_backend_type_by_name(self, backend_name: str) -> Dict[str, Any]:
        """ Gets the properties of a backend type, given the backend name (case insensitive).

        Args:
            backend_name: The backend name.

        Raises:
            ApiError: An ApiError exception is thrown when the backend name does not exist.

        Returns:
            The properties of the backend type of the specific backend.
            See `get_default_backend_type` for a description of the backend type properties.
        """
        backend_type = next((backend for backend in self.get_backend_types()
                            if backend['name'].lower() == backend_name.lower()), None)
        if backend_type is None:
            raise ApiError('Backend type with name {} does not exist!'.format(backend_name))
        return OrderedDict(backend_type)

    def get_backend_type(self, identifier: Optional[Union[int, str]] = None) -> Dict[str, Any]:
        """ Gets the properties of the backend type, given the identifier. If no identifier is given,
            the default backend type will be returned. With an identifier of type string or int,
            the backend type will be searched by name or id respectively.

        Args:
            identifier: The backend type identifier.

        Raises:
            ApiError: If the requested backend type does not exist.
            ValueError: If the backend type identifier is not of the correct type.

        Returns:
            This method returns the default backend type or the backend type as identified by
            the parameter of this method.
            See `get_default_backend_type` for a description of the backend type properties.
        """
        if identifier is None:
            return self.get_default_backend_type()
        elif isinstance(identifier, int):
            return self.get_backend_type_by_id(identifier)
        elif isinstance(identifier, str):
            return self.get_backend_type_by_name(identifier)
        else:
            raise ValueError('Identifier should be of type int, str or None!')

    #  projects  #

    def list_projects(self) -> None:
        """ Prints a list of all the projects registered to the current user the API is authenticated for.
            For each project the name, id and backend type is printed.
        """
        projects = self.get_projects()
        for project in projects:
            print(f'Project name: {project["name"]}, id: {project["id"]}, backend type: {project["backend_type"]}')

    def get_project(self, project_id: int) -> Dict[str, Any]:
        """ Gets the properties of a project, given the project id.
        Args:
            project_id: The project identification number.

        Raises:
            ApiError: If the requested project does not exist.

        Returns:
            The properties describing the project:
                | key                           | description
                |-------------------------------|----------------------------------------------------------------------
                | url (str)                     | The url for this project.
                | id (int)                      | Unique id of the project.
                | name (str)                    | Name of the project.
                | owner (int)                   | Url to get the owner of the project.
                | assets (str)                  | Url to get the assets of the project.
                | backend_type (str)            | Url to get the backend type of the project.
                | default_number_of_shots (int) | Default number of executions for this project.
        """
        try:
            project = self._action(['projects', 'read'], params={'id': project_id})
        except ErrorMessage as err_msg:
            raise ApiError('Project with id {} does not exist!'.format(project_id)) from err_msg
        return OrderedDict(project)

    def get_projects(self) -> List[Dict[str, Any]]:
        """ Gets all the projects registered to the user the API is currently authenticated for.

        Returns:
            The projects with all of its properties.
            See `get_project` for a description of the project properties.
        """
        ret: List[Dict[str, Any]] = self._action(['projects', 'list'])
        return ret

    def create_project(self, name: str, default_number_of_shots: int, backend_type: Dict[str, Any]) -> Dict[str, Any]:
        """ Creates a new project for executing cQASM code.

        Args:
            name: The name for the project. The name need not be unique.
            default_number_of_shots: The default number of executions of the program before collecting
                                     the results.
            backend_type: The properties of the backend type.

        Returns:
            The newly created project with all of its properties.
            See `get_project` for a description of the project properties.
        """
        payload = {
            'name': name,
            'default_number_of_shots': default_number_of_shots,
            'backend_type': backend_type['url'],
        }
        return OrderedDict((self._action(['projects', 'create'], params=payload)))

    def delete_project(self, project_id: int) -> None:
        """ Deletes the project identified by project_id together with all its assets, jobs and results.
            Only projects can be deleted that are registered for the user the API is currently authenticated for.

        Args:
            project_id: The project identification number.

        Raises:
            ApiError: If the project identified by project_id does not exist.
        """
        payload = {
            'id': project_id
        }
        try:
            self._action(['projects', 'delete'], params=payload)
        except ErrorMessage as err_msg:
            raise ApiError('Project with id {} does not exist!'.format(project_id)) from err_msg

    #  jobs  #

    def list_jobs(self) -> None:
        """ Prints a list of all the jobs registered to the current user the API is authenticated for.
            For each job the name, identification number and status is printed.
        """
        jobs = self.get_jobs()
        for job in jobs:
            print(f'Job name: {job["name"]}, id: {job["id"]}, status: {job["status"]}')

    def get_job(self, job_id: int) -> Dict[str, Any]:
        """ Gets the properties of a job, given the job id.
        Args:
            job_id: The job identification number.

        Raises:
            ApiError: If the requested job does not exist.

        Returns:
            The properties describing the job:
                | key                           | description
                |-------------------------------|----------------------------------------------------------------------
                | url (str)                     | The url for the job.
                | name (str)                    | Name of the circuit that is executed by this job.
                | id (int)                      | Unique id of the job.
                | status (str)                  | Execution status of the job: e.g. 'NEW', 'COMPLETE', 'CANCELLED',
                |                               | 'RUNNING'.
                | input (str)                   | Url to get the assets of the job.
                | backend (str)                 | Url to get the backend the job is executed on.
                | backend_type (str)            | Url to get the backend type of the backend the job is executed on.
                | results (str)                 | Url to get the results for the job.
                | queued_at (str)               | The date-time the job is queued at.
                |                               | The format is 'yyyy-MM-ddTHH:mm:ss.SSSSSSZ' Zulu Time.
                | number_of_shots (int)         | Number of executions for this job.
                | full_state_projection (bool)  | Indicates if the backend has full state projection.
                |                               | Used for optimizing simulations. For more information see:
                |                               | https://www.quantum-inspire.com/kbase/optimization-of-simulations/
                | user_data (str)               | The user configuration data.
        """
        try:
            job = self._action(['jobs', 'read'], params={'id': job_id})
        except ErrorMessage as err_msg:
            raise ApiError('Job with id {} does not exist!'.format(job_id)) from err_msg
        return OrderedDict(job)

    def get_jobs(self) -> List[Dict[str, Any]]:
        """ Gets all the jobs registered to projects for the user the API is currently authenticated for.

        Returns:
            The jobs with all of its properties.
            See `get_job` for a description of the job properties.
        """
        ret: List[Dict[str, Any]] = self._action(['jobs', 'list'])
        return ret

    def get_jobs_from_asset(self, asset_id: int) -> List[Dict[str, Any]]:
        """ Gets the jobs with its properties for an asset, given the asset id.

        Args:
            asset_id: The asset identification number.

        Returns:
            List of jobs with its properties for the asset with identification asset_id.
            An empty list is returned when the asset has no jobs.

        Raises:
            ApiError: If the asset identified by asset_id does not exist.
        """
        try:
            jobs = self._action(['assets', 'jobs', 'list'], params={'id': asset_id})
        except ErrorMessage as err_msg:
            raise ApiError('Asset with id {} does not exist!'.format(asset_id)) from err_msg
        ret: List[Dict[str, Any]] = jobs
        return ret

    def get_jobs_from_project(self, project_id: int) -> List[Dict[str, Any]]:
        """ Gets the jobs with its properties for a single project, given the project id.

        Args:
            project_id: The project identification number.

        Returns:
            List of jobs with its properties for the project with identification project_id.
            An empty list is returned when the project has no jobs.

        Raises:
            ApiError: If the project identified by project_id does not exist.
        """
        try:
            jobs = self._action(['projects', 'jobs', 'list'], params={'id': project_id})
        except ErrorMessage as err_msg:
            raise ApiError('Project with id {} does not exist!'.format(project_id)) from err_msg
        ret: List[Dict[str, Any]] = jobs
        return ret

    def delete_job(self, job_id: int) -> Dict[str, Any]:
        """ Deletes the job identified by job_id.
            Only jobs can be deleted that are registered for the user the API is currently authenticated for.

        Args:
            job_id: The job identification number.

        Returns:
            The deleted job indicated by the job identification number.
            See `get_job` for a description of the job properties.

        Raises:
            ApiError: If the job identified by job_id does not exist.
        """
        try:
            return OrderedDict(self._action(['jobs', 'delete'], params={'id': job_id}))
        except ErrorMessage as err_msg:
            raise ApiError('Job with id {} does not exist!'.format(job_id)) from err_msg

    def _create_job(self, name: str, asset: Dict[str, Any], project: Dict[str, Any], number_of_shots: int,
                    full_state_projection: bool = True, user_data: str = '') -> Dict[str, Any]:
        """ Creates a new job for executing cQASM code. This method is used by execute_qasm_async and indirectly
            by execute_qasm.

        Args:
            name: The name for the job.
            asset:  The asset with the cQASM code.
            project: The project with backend type.
            number_of_shots: The number of executions before returning the result.
            full_state_projection: Used for optimizing simulations. For more information see:
                                   https://www.quantum-inspire.com/kbase/optimization-of-simulations/
            user_data: Data that the user wants to pass along with the job.

        Returns:
            The properties describing the new job.
            See `get_job` for a description of the job properties.
        """
        payload = {
            'status': 'NEW',
            'name': name,
            'input': asset['url'],
            'backend_type': project['backend_type'],
            'number_of_shots': number_of_shots,
            'full_state_projection': full_state_projection,
            'user_data': user_data
        }
        return OrderedDict(self._action(['jobs', 'create'], params=payload))

    #  results  #

    def list_results(self) -> None:
        """ Prints a list of all the results registered to the current user the API is authenticated for.
            For each result the identification number and creation date is printed.
        """
        results = self.get_results()
        for result in results:
            print(f'Result id: {result["id"]} (date: {result["created_at"]})')

    def get_result(self, result_id: int) -> Dict[str, Any]:
        """ Gets the histogram results of the executed cQASM code, given the result_id.
        Args:
            result_id: The result identification number.

        Raises:
            ApiError: If the requested result does not exist.

        Returns:
            The properties describing the result:
                | key                               | description
                |-----------------------------------|-------------------------------------------------------------------
                | id (int)                          | Unique id of the result.
                | url (str)                         | The url to get this result.
                | job (str)                         | The url to get the job that generated the result.
                | created_at (str)                  | The date-time the result is created at.
                |                                   | The format is 'yyyy-MM-ddTHH:mm:ss.SSSSSSZ' Zulu Time.
                | number_of_qubits (int)            | Number of qubits in the circuit for this experiment.
                | execution_time_in_seconds (float) | The execution time of the job.
                | raw_text (str)                    | Text string filled when an error occurred, else empty.
                | raw_data_url (str)                | Url to get the raw data of the result. The raw data exists of a
                |                                   | list of integer values depicting the state for each shot.
                | histogram (OrderedDict)           | The histogram as a list of tuples with state (str) and
                |                                   | its probability (float).
                | histogram_url (str)               | Url to get the histogram with probabilities. This results in the
                |                                   | OrderedDict as found in property histogram of result.
                | measurement_mask (int)            | (deprecated, unused) The measurement mask.
                | quantum_states_url (str)          | Url to get a list of quantum states.
                | measurement_register_url (str)    | Url to get a list of measurement register.
        """
        try:
            result = self._action(['results', 'read'], params={'id': result_id})
        except ErrorMessage as err_msg:
            raise ApiError('Result with id {} does not exist!'.format(result_id)) from err_msg
        return OrderedDict(result)

    def get_results(self) -> List[Dict[str, Any]]:
        """ Gets all the results registered for the user the API is currently authenticated for.

        Returns:
            The results with all of its properties.
            See `get_result` for a description of the result properties.
        """
        ret: List[Dict[str, Any]] = self._action(['results', 'list'])
        return ret

    def get_result_from_job(self, job_id: int) -> Dict[str, Any]:
        """ Gets the result with its properties for a single job, given the job id.

        Args:
            job_id: The job identification number.

        Returns:
            The result with its properties for the job with identification job_id.
            See `get_result` for a description of the result properties.

        Raises:
            ApiError: If the job identified by job_id does not exist.
        """
        try:
            result = self._action(['jobs', 'result', 'list'], params={'id': job_id})
        except ErrorMessage as err_msg:
            raise ApiError('Job with id {} does not exist!'.format(job_id)) from err_msg
        return OrderedDict(result)

    def get_raw_data_from_result(self, result_id: int) -> List[int]:
        """ Gets the raw data from the result of the executed cQASM code, given the result_id. The raw data consists
            of a list with integer state values for each shot of the experiment (see job.number_of_shots).

        Args:
            result_id: The identification number of the result.

        Raises:
            ApiError: If the raw data url in result is invalid or the request for the raw data using the url failed.

        Returns:
            The raw data as a list of integer values. An empty list is returned when there is no raw data.
        """
        result = self.get_result(result_id)
        raw_data_url = str(result.get('raw_data_url'))
        try:
            token = raw_data_url.split('/')[-2]
        except IndexError as err_msg:
            raise ApiError('Invalid raw data url for result with id {}!'.format(result_id)) from err_msg
        try:
            raw_data: List[int] = self._action(['results', 'raw-data', 'read'], params={'id': result_id,
                                                                                        'token': token})
        except ErrorMessage as err_msg:
            raise ApiError('Raw data for result with id {} does not exist!'.format(result_id)) from err_msg
        return raw_data

    def get_quantum_states_from_result(self, result_id: int) -> List[Any]:
        """ Gets the quantum states from the result of the executed cQASM code, given the result_id.

        Args:
            result_id: The identification number of the result.

        Raises:
            ApiError: If the quantum states url in result is invalid or the request for the quantum states using the
            url failed.

        Returns:
            The quantum states consists of a list of quantum state values. An empty list is returned when there is
            no data.
        """
        result = self.get_result(result_id)
        quantum_states_url = str(result.get('quantum_states_url'))
        try:
            token = quantum_states_url.split('/')[-2]
        except IndexError as err_msg:
            raise ApiError('Invalid quantum states url for result with id {}!'.format(result_id)) from err_msg
        try:
            quantum_states: List[Any] = self._action(['results', 'quantum-states', 'read'], params={'id': result_id,
                                                                                                    'token': token})
        except ErrorMessage as err_msg:
            raise ApiError('Quantum states for result with id {} does not exist!'.format(result_id)) from err_msg
        return quantum_states

    def get_measurement_register_from_result(self, result_id: int) -> List[Any]:
        """ Gets the measurement register from the result of the executed cQASM code, given the result_id.

        Args:
            result_id: The identification number of the result.

        Raises:
            ApiError: If the measurement register url in result is invalid or the request for the measurement register
            using the url failed.

        Returns:
            The measurement register consists of a list of measurement register values. An empty list is returned
            when there is no data.
        """
        result = self.get_result(result_id)
        measurement_register_url = str(result.get('measurement_register_url'))
        try:
            token = measurement_register_url.split('/')[-2]
        except IndexError as err_msg:
            raise ApiError('Invalid measurement register url for result with id {}!'.format(result_id)) from err_msg
        try:
            measurement_register: List[Any] = self._action(['results', 'measurement-register', 'read'],
                                                           params={'id': result_id,
                                                                   'token': token})
        except ErrorMessage as err_msg:
            raise ApiError('Measurement register for result with id {} does not exist!'.format(result_id)) from err_msg
        return measurement_register

    #  assets  #

    def list_assets(self) -> None:
        """ Prints a list of the assets registered to the current user the API is authenticated for.
            For each asset the name, identification number and project identification number is printed."""
        assets = self.get_assets()
        for asset in assets:
            print(f'Asset name: {asset["name"]}, id: {asset["id"]}, (project_id: {asset["project_id"]})')

    def get_asset(self, asset_id: int) -> Dict[str, Any]:
        """ Gets the properties of the asset, given the asset_id.
        Args:
            asset_id: The asset identification number.

        Raises:
            ApiError: If the requested asset does not exist.

        Returns:
            The properties describing the asset:
                | key                       | description
                |---------------------------|---------------------------------------------------------------------------
                | url (str)                 | The url to get this asset.
                | id (int)                  | Unique id of this asset.
                | name (str)                | The name to get the asset.
                | contentType (str)         | The description of the content e.g. 'application/qasm' or
                |                           | 'text/plain'.
                | content (str)             | The content itself. For example a cQASM program when linked to a job.
                | project (str)             | Url to get the project properties for which this asset was created.
                | project_id (int)          | The project id of the project for which this asset was created.
        """
        try:
            asset = self._action(['assets', 'read'], params={'id': asset_id})
        except ErrorMessage as err_msg:
            raise ApiError('Asset with id {} does not exist!'.format(asset_id)) from err_msg
        return OrderedDict(asset)

    def get_assets(self) -> List[Dict[str, Any]]:
        """ Gets all the assets registered for the user the API is currently authenticated for.

        Returns:
            The assets with all of its properties.
            See `get_asset` for a description of the asset properties.
        """
        ret: List[Dict[str, Any]] = self._action(['assets', 'list'])
        return ret

    def get_assets_from_project(self, project_id: int) -> List[Dict[str, Any]]:
        """ Gets the assets with its properties for a single project, given the project id.

        Args:
            project_id: The project identification number.

        Returns:
            List of assets with its properties for the project with identification project_id.
            An empty list is returned when the project has no assets.

        Raises:
            ApiError: If the project identified by project_id does not exist.
        """
        try:
            assets: List[Dict[str, Any]] = self._action(['projects', 'assets', 'list'], params={'id': project_id})
        except ErrorMessage as err_msg:
            raise ApiError('Project with id {} does not exist!'.format(project_id)) from err_msg
        return assets

    def get_asset_from_job(self, job_id: int) -> Dict[str, Any]:
        """ Gets the asset data from the job, given the job_id.

        Args:
            job_id: The identification number of the job.

        Raises:
            ApiError: If the requested asset with identification from the job 'input' field does not exist.

        Returns:
            The assets with all of its properties.
            See `get_asset` for a description of the asset properties.
        """
        job = self.get_job(job_id)
        asset_url = str(job.get('input'))
        try:
            asset_id = int(asset_url.split('/')[-2])
        except (ValueError, IndexError) as err_msg:
            raise ApiError('Invalid input url for job with id {}!'.format(job_id)) from err_msg
        try:
            asset = self._action(['assets', 'read'], params={'id': asset_id})
        except ErrorMessage as err_msg:
            raise ApiError('Asset with id {} does not exist!'.format(asset_id)) from err_msg
        return OrderedDict(asset)

    def _create_asset(self, name: str, project: Dict[str, Any], content: str) -> Dict[str, Any]:
        """ This method is used by execute_qasm_async to generate a new asset with a unique name to hold the
            content of the cQASM program for the project given. Assets are deleted when the project for which the asset
            was created is deleted.

        Args:
            name: The name of the new asset.
            project: The project to which the asset is linked to.
            content: The cQASM code content.

        Returns:
            The properties of the asset.
            See `get_asset` for a description of the asset properties.
        """
        payload = {
            'name': name,
            'contentType': 'application/qasm',
            'project': project['url'],
            'content': content,
        }
        return OrderedDict(self._action(['assets', 'create'], params=payload))

    #  other  #

    @staticmethod
    def _wait_for_completed_job(quantum_inspire_job: QuantumInspireJob, collect_max_tries: Optional[int],
                                sec_retry_delay: float = 0.5) -> bool:
        """ Holds the process and requests the job status until completed or when
            the maximum number of tries has been reached.

        Args:
            quantum_inspire_job: A job object.
            collect_max_tries: The maximum number of request tries.
            sec_retry_delay: The time delay in between job status checks in seconds.

        Returns:
            True if the job result could be collected else False.
        """
        attempts = itertools.count() if collect_max_tries is None else range(collect_max_tries)
        for _ in attempts:
            time.sleep(sec_retry_delay)
            if quantum_inspire_job.check_status() == 'COMPLETE':
                return True
        return False

    def execute_qasm(self, qasm: str, backend_type: Optional[Union[Dict[str, Any], int, str]] = None,
                     number_of_shots: int = 256, collect_tries: Optional[int] = None,
                     default_number_of_shots: int = 256, identifier: Optional[str] = None,
                     full_state_projection: bool = True) -> Dict[str, Any]:
        """ With this method a cQASM program is executed, and the result is returned when the job is completed.

            The method 'execute_qasm_async' is called which returns a QuantumInspireJob directly without waiting
            for the job to complete. After this call a waiting loop is started to wait for the job to finish and get
            the result.
            When no project name was given when the QuantumInspireAPI was created, the job is linked to a newly
            created temporary project in 'execute_qasm_async'. When the job has finished running, this project is
            deleted.

            Depending on how busy the backend is, it takes some time to execute the job and
            returning the result. This method waits for the job to finish. The parameter collect_tries defines the
            maximum waiting time (collect_tries x 0.5 seconds). When the job takes longer to finish no results
            are returned.

        Args:
            qasm: The cQASM code as string object.
            backend_type: The backend_type to execute the algorithm on.
            number_of_shots: Execution times of the algorithm before collecting the results.
            collect_tries: The number of times the status of the job is check for completion before returning.
            default_number_of_shots: The default used number of shots for the project.
            identifier: The identifier to generate names for the project, asset and job when necessary.
            full_state_projection: Do not use full state projection when set to False (default is True).

        Returns:
            The results of the executed cQASM if successful else an empty dictionary if
            the results could not be collected within the given number of tries.
            See `get_result` for a description of the result properties.
        """
        delete_project_afterwards = self.project_name is None
        quantum_inspire_job = None
        try:
            quantum_inspire_job = self.execute_qasm_async(qasm, backend_type=backend_type,
                                                          number_of_shots=number_of_shots,
                                                          default_number_of_shots=default_number_of_shots,
                                                          identifier=identifier,
                                                          full_state_projection=full_state_projection)

            has_results = self._wait_for_completed_job(quantum_inspire_job, collect_tries)
            return OrderedDict(quantum_inspire_job.retrieve_results()) if has_results else OrderedDict()
        finally:
            if delete_project_afterwards and quantum_inspire_job is not None:
                project_identifier = quantum_inspire_job.get_project_identifier()
                self.delete_project(project_identifier)

    def execute_qasm_async(self, qasm: str, backend_type: Optional[Union[Dict[str, Any], int, str]] = None,
                           number_of_shots: int = 256, default_number_of_shots: int = 256,
                           identifier: Optional[str] = None, full_state_projection: bool = True,
                           project: Optional[Dict[str, Any]] = None, job_name: Optional[str] = None,
                           user_data: str = '') -> QuantumInspireJob:
        """ With this method a cQASM program (job) is scheduled to be executed asynchronously. The method returns
            directly without waiting for the job to complete, as opposed to method `execute_qasm` which waits for
            the job to finish and returns the result.

            To execute a cQASM program a job is scheduled on a backend of type 'backend type' as given by the
            parameter backend_type. Currently there are 3 backend types available:
            1. 'QX single-node simulator'          (default for anonymous accounts)
            2. 'QX single-node simulator SurfSara' (advanced account credentials needed)
            3. 'QX multi-node simulator SurfSara'  (advanced account credentials needed)
            When no backend_type is given, the default backend type currently being 'QX single-node simulator', is used.

            The job has to be linked with a project before it can be scheduled to execute.

            When a project name was supplied when the QuantumInspireAPI was created (see __init__), the job
            is linked to an existing project with this project name. When no project exists with this name a
            project with this name is created.
            When no project name was given when the QuantumInspireAPI was created, the job is linked to the project
            that is given as an argument 'project'. When this 'project' argument is empty, a project is created.
            First a project name is generated using the parameter 'identifier' or when parameter 'identifier' is
            empty an identifier is generated.

            When the project is created, it is created for the backend type and the default number of shots given.
            When the project already existed the values of the existing project are used for backend type and
            default number of shots.

            An asset with a unique id is created containing the cQASM program. This asset is linked to the project.

            When the project and the asset containing the program are known, a job is created with the name given by
            parameter job_name. When this parameter job_name is not filled, a job name is generated.
            The job that is created for running the program (contained in the asset) is linked to the project and will
            be executed number_of_shots times (as given by the parameter) before the results can be collected.
            The jobs' user_data is filled with the user data given as a parameter. This user data can be fetched and
            used later in the process. The job full_state_projection defaults to True because Quantum Inspire uses
            full_state_projection. When set to False, the job does not use full state projection.

            Once the job is created, the method returns directly without waiting for the job to complete.
            The job is returned as a QuantumInspireJob. This class encapsulates the job and contains methods the get
            the status of the job and retrieve the execution results when the job is completed.

        Args:
            qasm: The qasm code as a string object.
            backend_type: The backend_type to execute the algorithm on.
            number_of_shots: Execution times of the algorithm before the results can be collected.
            default_number_of_shots: The default used number of shots for the project.
            identifier: The identifier used for generating names for the project, asset and job.
            full_state_projection: Do not use full state projection when set to False (default is True).
            project: The properties of an existing project, the asset and job are linked to. Only used
                     when the project_name member of the api is empty.
            job_name: Name for the job that is to be executed, when None a job name is generated (see identifier)
            user_data: Data that the user wants to pass along with the job.

        Returns:
            An encapsulated job object containing methods the get the status of the job and
            retrieve the execution results.
        """
        if not isinstance(backend_type, OrderedDict):
            if backend_type is None:
                backend_type = self.get_backend_type(None)
            elif isinstance(backend_type, int):
                backend_type = self.get_backend_type(int(backend_type))
            elif isinstance(backend_type, str):
                backend_type = self.get_backend_type(str(backend_type))

        if identifier is None:
            identifier = str(uuid.uuid1())

        if self.project_name is not None:
            project = next((project for project in self.get_projects()
                            if project['name'] == self.project_name), None)

        if project is None:
            project_name = self.project_name if self.project_name else 'qi-sdk-project-{}'.format(identifier)
            project = self.create_project(project_name, default_number_of_shots, backend_type)

        asset_name = 'qi-sdk-asset-{}'.format(identifier)
        asset = self._create_asset(asset_name, project, qasm)

        if job_name is None:
            job_name = 'qi-sdk-job-{}'.format(identifier)
        job = self._create_job(job_name, asset, project, number_of_shots, user_data=user_data,
                               full_state_projection=full_state_projection)

        return QuantumInspireJob(self, job['id'])



from typing import Dict, Any
from coreapi.exceptions import ErrorMessage


class QuantumInspireJob:

    def __init__(self, api: Any, job_identifier: int) -> None:
        """ The QuantumInspire Job class encapsulates the base job of the API and has
            methods to check the status and retrieve the results from the API.

        Arguments:
            api: An instance to the API.
            job_identifier: The job identification number.
        """
        QuantumInspireJob.__check_arguments(api, job_identifier)
        self.__job_identifier: int = job_identifier
        self.__api: Any = api

    @staticmethod
    def __check_arguments(api: Any, job_identifier: int) -> None:
        """ Checks whether the supplied arguments are of correct type.

        Arguments:
            api: An instance to the API.
            job_identifier: The job identification number.

        Raises:
            ValueError: When the api is not a QuantumInspireApi or when the
            job identifier is not found.
        """
        if type(api).__name__ != 'QuantumInspireAPI':
            raise ValueError('Invalid Quantum Inspire API!')
        try:
            _ = api.get_job(job_identifier)
        except ErrorMessage as error:
            raise ValueError('Invalid job identifier!') from error

    def check_status(self) -> str:
        """ Checks the execution status of the job.

        Returns:
            The status of the job.
        """
        job = self.__api.get_job(self.__job_identifier)
        return str(job['status'])

    def retrieve_results(self) -> Dict[str, Any]:
        """ Gets the results of the job.

        Returns:
            The execution results with a histogram item containing the result
            histogram of the job. When an error has occurred the raw_text item shall not be
            an empty string.
        """
        result: Dict[str, Any] = self.__api.get_result_from_job(self.__job_identifier)
        return result

    def get_job_identifier(self) -> int:
        """ Gets the set job identification number for the wrapped job.

        Returns:
            The job identification number.
        """
        return self.__job_identifier

    def get_project_identifier(self) -> int:
        """ Gets the project identification number of the wrapped job.

        Returns:
            The project identification number.
        """
        asset = self.__api.get_asset_from_job(self.__job_identifier)
        return int(asset['project_id'])



""" Quantum Inspire SDK

Copyright 2018 QuTech Delft

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""
from setuptools import setup


def get_version_number(module):
    """ Extract the version number from the source code.

        Returns:
            Str: the version number.
    """
    with open('src/{}/version.py'.format(module), 'r') as file_stream:
        line = file_stream.readline().split()
        version_number = line[2].replace('\'', '')
    return version_number


def get_long_description():
    """ Extract the long description from the README file """

    with open('README.md', encoding='utf-8') as f:
        long_description = f.read()

    return long_description


setup(name='quantuminspire',
      description='SDK for the Quantum Inspire platform',
      long_description=get_long_description(),
      long_description_content_type='text/markdown',
      version=get_version_number('quantuminspire'),
      author='QuantumInspire',
      python_requires='>=3.6',
      package_dir={'': 'src'},
      data_files=[('share/doc/quantuminspire/examples', ['docs/example_qiskit_entangle.py',
                                                         'docs/example_projectq_entangle.py',
                                                         'docs/example_projectq_grover.py',
                                                         'docs/example_projectq.ipynb',
                                                         'docs/grover_algorithm_qi.ipynb',
                                                         'docs/qi-performance-test.ipynb',
                                                         'docs/grover-qi.png'])],
      classifiers=[
          'Development Status :: 4 - Beta',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'License :: OSI Approved :: Apache Software License'],
      license='Apache 2.0',
      packages=['quantuminspire', 'quantuminspire.qiskit', 'quantuminspire.projectq'],
      install_requires=['pytest>=3.3.1', 'coverage>=4.5.1', 'matplotlib>=2.1',
                        'coreapi>=2.3.3', 'numpy', 'jupyter'],
      extras_require={'qiskit': ["qiskit>=0.7.0, <0.9.0"], 'projectq': ["projectq>=0.4"]})



""" Quantum Inspire SDK

Copyright 2018 QuTech Delft

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""
import io
import json
import uuid
import numpy as np
from typing import Dict, List, Tuple, Optional, Any
from collections import defaultdict, OrderedDict, Counter

from coreapi.exceptions import ErrorMessage
from qiskit.providers import BaseBackend
from qiskit.providers.models import BackendConfiguration
from qiskit.providers.models.backendconfiguration import GateConfig
from qiskit.qobj import Qobj, QobjExperiment
from qiskit.result.models import ExperimentResult, ExperimentResultData
from qiskit.validation.base import Obj

from quantuminspire.qiskit.circuit_parser import CircuitToString
from quantuminspire.qiskit.qi_job import QIJob
from quantuminspire.api import QuantumInspireAPI
from quantuminspire.exceptions import QisKitBackendError
from quantuminspire.job import QuantumInspireJob
from quantuminspire.version import __version__ as quantum_inspire_version


class QuantumInspireBackend(BaseBackend):  # type: ignore
    DEFAULT_CONFIGURATION = BackendConfiguration(
        backend_name='qi_simulator',
        backend_version=quantum_inspire_version,
        n_qubits=26,
        basis_gates=['x', 'y', 'z', 'h', 'rx', 'ry', 'rz', 's', 'sdg', 't', 'tdg', 'cx', 'ccx', 'u1', 'u2', 'u3', 'id',
                     'swap', 'snapshot'],
        gates=[GateConfig(name='NotUsed', parameters=['NaN'], qasm_def='NaN')],
        local=False,
        simulator=True,
        conditional=True,
        open_pulse=False,
        memory=True,
        max_shots=1024
    )

    def __init__(self, api: QuantumInspireAPI, provider: Any,
                 configuration: Optional[BackendConfiguration] = None) -> None:
        """ Python implementation of a quantum simulator using Quantum Inspire API.

        Args:
            api: The interface instance to the Quantum Inspire API.
            provider: Provider for this backend.
            configuration: The configuration of the quantum inspire backend. The
                configuration must implement the fields given by the QiSimulatorPy.DEFAULT_CONFIGURATION. All
                configuration fields are listed in the table below. The table rows with an asterisk specify fields which
                can have a custom value and are allowed to be changed according to the description column.

                | key                     | description
                |-------------------------|----------------------------------------------------------------------------
                | backend_name (str)*     | The name of the quantum inspire backend. The API can list the name of each
                |                         | available backend using the function api.list_backend_types(). One of the
                |                         | listed names must be used.
                | backend_version (str)   | Backend version in the form X.Y.Z.
                | n_qubits (int)          | Number of qubits.
                | basis_gates (list[str]) | A list of basis gates to compile to.
                | gates (GateConfig)      | List of basis gates on the backend. Not used.
                | local (bool)            | Indicates whether the system is running locally or remotely. Not used.
                | simulator (bool)        | Specifies whether the backend is a simulator or a quantum system. Not used.
                | conditional (bool)      | Backend supports conditional operations.
                | open_pulse (bool)       | Backend supports open pulse. False.
                | memory (bool)           | Backend supports memory. True.
                | max_shots (int)         | Maximum number of shots supported.
        """
        super().__init__(configuration=(configuration or
                                        QuantumInspireBackend.DEFAULT_CONFIGURATION),
                         provider=provider)
        self.__backend: Dict[str, Any] = api.get_backend_type_by_name(self.name())
        self.__api: QuantumInspireAPI = api

    @property
    def backend_name(self) -> str:
        return self.name()  # type: ignore

    def run(self, qobj: Qobj) -> QIJob:
        """ Submits a quantum job to the Quantum Inspire platform.

        Args:
            qobj: The quantum job with the Qiskit algorithm and quantum inspire backend.

        Returns:
            A job that has been submitted.
        """
        self.__validate(qobj)
        number_of_shots = qobj.config.shots

        identifier = uuid.uuid1()
        project_name = 'qi-sdk-project-{}'.format(identifier)
        project = self.__api.create_project(project_name, number_of_shots, self.__backend)
        experiments = qobj.experiments
        job = QIJob(self, str(project['id']), self.__api)
        [self._submit_experiment(experiment, number_of_shots, project=project) for experiment in experiments]
        job.experiments = experiments
        return job

    def retrieve_job(self, job_id: str) -> QIJob:
        """ Retrieve a specified job by its job_id.

        Args:
            job_id: The job id.

        Returns:
            The job that has been retrieved.

        Raises:
            QisKitBackendError: If job not found or error occurs during retrieval of the job.
        """
        try:
            self.__api.get_project(int(job_id))
        except (ErrorMessage, ValueError):
            raise QisKitBackendError("Could not retrieve job with job_id '{}' ".format(job_id))
        return QIJob(self, job_id, self.__api)

    @staticmethod
    def _generate_cqasm(experiment: QobjExperiment) -> str:
        """ Generates the cQASM from the Qiskit experiment.

        Args:
            experiment: The experiment that contains instructions to be converted to cQASM.

        Returns:
            The cQASM code that can be sent to the Quantum Inspire API.
        """
        parser = CircuitToString()
        number_of_qubits = experiment.header.n_qubits
        instructions = experiment.instructions
        with io.StringIO() as stream:
            stream.write('version 1.0\n')
            stream.write('# cQASM generated by QI backend for Qiskit\n')
            stream.write('qubits %d\n' % number_of_qubits)
            for instruction in instructions:
                parser.parse(stream, instruction)
            return stream.getvalue()

    def _submit_experiment(self, experiment: QobjExperiment, number_of_shots: int,
                           project: Optional[Dict[str, Any]] = None) -> QuantumInspireJob:
        compiled_qasm = self._generate_cqasm(experiment)
        measurements = self._collect_measurements(experiment)
        user_data = {'name': experiment.header.name, 'memory_slots': experiment.header.memory_slots,
                     'creg_sizes': experiment.header.creg_sizes, 'measurements': measurements}
        job_id = self.__api.execute_qasm_async(compiled_qasm, backend_type=self.__backend,
                                               number_of_shots=number_of_shots, project=project,
                                               job_name=experiment.header.name, user_data=json.dumps(user_data))
        return job_id

    def get_experiment_results(self, qi_job: QIJob) -> List[ExperimentResult]:
        """ Get results from experiments from the Quantum-inspire platform.

        Args:
            qi_job: A job that has already been submitted and which execution is completed.

        Raises:
            QisKitBackendError: If an error occurred during execution by the backend.

        Returns:
            A list of experiment results; containing the data, execution time, status, etc.
        """
        jobs = self.__api.get_jobs_from_project(int(qi_job.job_id()))
        results = [self.__api.get_result_from_job(job['id']) for job in jobs]
        experiment_results = []
        for result, job in zip(results, jobs):
            if not result.get('histogram', {}):
                raise QisKitBackendError(
                    'Result from backend contains no histogram data!\n{}'.format(result.get('raw_text')))

            user_data = json.loads(str(job.get('user_data')))
            measurements = user_data.pop('measurements')
            histogram_obj, memory_data = self.__convert_result_data(result, measurements)
            full_state_histogram_obj = self.__convert_histogram(result, measurements)
            experiment_result_data = ExperimentResultData(counts=histogram_obj,
                                                          probabilities=full_state_histogram_obj,
                                                          memory=memory_data)
            header = Obj.from_dict(user_data)
            experiment_result_dictionary = {'name': job.get('name'), 'seed': 42, 'shots': job.get('number_of_shots'),
                                            'data': experiment_result_data, 'status': 'DONE', 'success': True,
                                            'time_taken': result.get('execution_time_in_seconds'), 'header': header}
            experiment_results.append(ExperimentResult(**experiment_result_dictionary))
        return experiment_results

    def __validate(self, job: Qobj) -> None:
        """ Validates the number of shots, classical bits and compiled Qiskit circuits.

        Args:
            job: The quantum job with the Qiskit algorithm and quantum inspire backend.
        """
        QuantumInspireBackend.__validate_number_of_shots(job)

        for experiment in job.experiments:
            self.__validate_number_of_clbits(experiment)
            QuantumInspireBackend.__validate_no_gates_after_measure(experiment)

    @staticmethod
    def __validate_number_of_shots(job: Qobj) -> None:
        """ Checks whether the number of shots has a valid value.

        Args:
            job: The quantum job with the Qiskit algorithm and quantum inspire backend.

        Raises:
            QisKitBackendError: When the value is not correct.
        """
        number_of_shots = job.config.shots
        if number_of_shots < 1:
            raise QisKitBackendError('Invalid shots (number_of_shots={})'.format(number_of_shots))

    def __validate_number_of_clbits(self, experiment: QobjExperiment) -> None:
        """ Checks whether the number of classical bits has a valid value.

        Args:
            experiment: The experiment with gate operations and header.

        Raises:
            QisKitBackendError: When the value is not correct.
        """
        number_of_clbits = experiment.header.memory_slots
        if number_of_clbits < 1:
            raise QisKitBackendError("Invalid amount of classical bits ({})!".format(number_of_clbits))

        if BaseBackend.configuration(self).conditional:
            number_of_qubits = experiment.header.n_qubits
            if number_of_clbits > number_of_qubits:
                # no problem when there are no conditional gate operations
                for instruction in experiment.instructions:
                    if hasattr(instruction, 'conditional'):
                        raise QisKitBackendError("Number of classical bits must be less than or equal to the"
                                                 " number of qubits when using conditional gate operations")

    @staticmethod
    def __validate_no_gates_after_measure(experiment: QobjExperiment) -> None:
        """ Checks whether the number of classical bits has a valid value.

        Args:
            experiment: The experiment with gate operations and header.

        Raises:
            QisKitBackendError: When the value is not correct.
        """
        measured_qubits = []
        for instruction in experiment.instructions:
            for qubit in instruction.qubits:
                if instruction.name == 'measure':
                    measured_qubits.append(qubit)
                elif qubit in measured_qubits:
                    raise QisKitBackendError('Operation on qubit {} after measurement'.format(qubit))

    @staticmethod
    def _collect_measurements(experiment: QobjExperiment) -> Dict[str, Any]:
        """ Determines the measured qubits and classical bits. The full-state measured
            qubits is returned when no measurements are present in the compiled circuit.

        Args:
            experiment: The experiment with gate operations and header.

        Returns:
            The dict contains measurements, which is a list of lists, for each measurement the list contains
            a list of [qubit_index, classical_bit_index], which represents the measurement of a qubit to a
            classical bit, and the second field in the dict is the number of classical bits (int).
        """
        header = experiment.header
        number_of_qubits = header.n_qubits
        number_of_clbits = header.memory_slots

        operations = experiment.instructions
        measurements = [[number_of_qubits - 1 - m.qubits[0],
                         number_of_clbits - 1 - m.memory[0]]
                        for m in operations if m.name == 'measure']
        if not measurements:
            measurements = [[index, index] for index in range(number_of_qubits)]
        return {'measurements': measurements, 'number_of_clbits': number_of_clbits}

    @staticmethod
    def __qubit_to_classical_hex(qubit_register: str, measurements: Dict[str, Any], number_of_qubits: int) -> str:
        """ This function converts the qubit register data to the hexadecimal representation of the classical state.

        Args:
            qubit_register: The measured value of the qubits represented as int.
            measurements: The dictionary contains a measured qubits/classical bits map (list) and the
                          number of classical bits (int).
            number_of_qubits: Number of qubits used in the algorithm.

        Returns:
            The hexadecimal value of the classical state.
        """
        qubit_state = ('{0:0{1}b}'.format(int(qubit_register), number_of_qubits))
        classical_state = ['0'] * measurements['number_of_clbits']
        for q, c in measurements['measurements']:
            classical_state[c] = qubit_state[q]
        classical_state_str = ''.join(classical_state)
        classical_state_hex = hex(int(classical_state_str, 2))
        return classical_state_hex

    @staticmethod
    def __convert_histogram(result: Dict[str, Any], measurements: Dict[str, Any]) -> Obj:
        """ The quantum inspire backend always uses full state projection. The SDK user
            can measure not all qubits and change the combined classical bits. This function
            converts the result to a histogram output that represents the probabilities
            measured with the classical bits.

        Args:
            result: The result output from the quantum inspire backend with full-
                    state projection histogram output.
            measurements: The dictionary contains a measured qubits/classical bits map (list) and the
                          number of classical bits (int).

        Returns:
            The resulting full state histogram with probabilities.
        """
        output_histogram_probabilities: Dict[str, float] = defaultdict(lambda: 0)
        number_of_qubits = result['number_of_qubits']
        state_probability: Dict[str, float] = result['histogram']
        for qubit_register, probability in state_probability.items():
            classical_state_hex = QuantumInspireBackend.__qubit_to_classical_hex(qubit_register, measurements,
                                                                                 number_of_qubits)
            output_histogram_probabilities[classical_state_hex] += probability

        sorted_histogram_probabilities: List[Tuple[Any, Any]] = sorted(output_histogram_probabilities.items(),
                                                                       key=lambda kv: int(kv[0], 16))
        full_state_histogram_obj = OrderedDict(sorted_histogram_probabilities)
        return Obj.from_dict(full_state_histogram_obj)

    def __convert_result_data(self, result: Dict[str, Any], measurements: Dict[str, Any]) -> Tuple[Obj, List[str]]:
        """ The quantum inspire backend returns the single shot values as raw data. This function
            converts this list of single shot values to hexadecimal memory data according the Qiskit spec.
            From this memory data the counts histogram is constructed by counting the single shot values.

        Note:
            When shots = 1, the backend returns an empty list as raw_data. This is a special case. In this case the
            resulting memory data consists of 1 value and the count histogram consists of 1 instance of this value.
            To determine this value a random float is generated in the range [0, 1). With this random number the
            value from this probabilities histogram is taken where the added probabilities is greater this random
            number.
            Example: probability histogram is {[0x0, 0.2], [0x3, 0.4], [0x5, 0.1], [0x6, 0.3]}.
            When random is in the range [0, 0.2) the first value of the probability histogram is taken (0x0).
            When random is in the range [0.2, 0.6) the second value of the probability histogram is taken (0x3).
            When random is in the range [0.6, 0.7) the third value of the probability histogram is taken (0x5).
            When random is in the range [0.7, 1) the last value of the probability histogram is taken (0x6).

        Args:
            result: The result output from the quantum inspire backend with full-
                    state projection histogram output.
            measurements: The dictionary contains a measured qubits/classical bits map (list) and the
                          number of classical bits (int).

        Returns:
            The result consists of two formats for the result. The first result is the histogram with count data,
            the second result is a list with converted hexadecimal memory values for each shot.
        """
        memory_data = []
        histogram_data: Dict[str, int] = defaultdict(lambda: 0)
        number_of_qubits: int = result['number_of_qubits']
        raw_data = self.__api.get_raw_data_from_result(result['id'])
        if raw_data:
            for raw_qubit_register in raw_data:
                classical_state_hex = QuantumInspireBackend.__qubit_to_classical_hex(str(raw_qubit_register),
                                                                                     measurements, number_of_qubits)
                memory_data.append(classical_state_hex)
            histogram_data = {elem: count for elem, count in Counter(memory_data).items()}
        else:
            state_probabilities = result['histogram']
            random_probability = np.random.rand()
            sum_probability = 0.0
            for qubit_register, probability in state_probabilities.items():
                sum_probability += probability
                if random_probability < sum_probability:
                    classical_state_hex = QuantumInspireBackend.__qubit_to_classical_hex(qubit_register, measurements,
                                                                                         number_of_qubits)
                    memory_data.append(classical_state_hex)
                    histogram_data[classical_state_hex] = 1
                    break

        sorted_histogram_data: List[Tuple[Any, Any]] = sorted(histogram_data.items(), key=lambda kv: int(kv[0], 16))
        histogram_obj = OrderedDict(sorted_histogram_data)
        return Obj.from_dict(histogram_obj), memory_data
